import spacy

from autogoal.grammar import Categorical, Boolean
from autogoal.kb import Sentence, Tuple, Word, Flags, List
from autogoal.utils import nice_repr


@nice_repr
class SpacyNLP:
    def __init__(
        self,
        language: Categorical("en", "es"),
        extract_pos: Boolean(),
        extract_lemma: Boolean(),
        extract_pos_tag: Boolean(),
        extract_dep: Boolean(),
        extract_entity: Boolean(),
        extract_details: Boolean(),
        extract_sentiment: Boolean(),
    ):
        self.language = language
        self.extract_pos = extract_pos
        self.extract_lemma = extract_lemma
        self.extract_pos_tag = extract_pos_tag
        self.extract_dep = extract_dep
        self.extract_entity = extract_entity
        self.extract_details = extract_details
        self.extract_sentiment = extract_sentiment
        self._nlp = None

    @property
    def nlp(self):
        if self._nlp is None:
            self._nlp = spacy.load(self.language)

        return self._nlp

    def run(self, input: Sentence()) -> Tuple(List(Word()), List(Flags())):
        tokenized = self.nlp(input)

        tokens = []
        flags = []

        for token in tokenized:
            token_flags = {}
            if self.extract_lemma:
                token_flags["lemma"] = token.lemma_
            if self.extract_pos_tag:
                token_flags["pos"] = token.pos_

                for kv in token.tag_.split("|"):
                    kv = kv.split("=")
                    if len(kv) == 2:
                        token_flags["tag_" + kv[0]] = kv[1]
                    else:
                        token_flags["tag_" + kv[0]] = True

            if self.extract_dep:
                token_flags["dep"] = token.dep_
            if self.extract_entity:
                token_flags["ent_type"] = token.ent_type_
                token_flags["ent_kb_id"] = token.ent_kb_id_
            if self.extract_details:
                token_flags["is_alpha"] = token.is_alpha
                token_flags["is_ascii"] = token.is_ascii
                token_flags["is_digit"] = token.is_digit
                token_flags["is_lower"] = token.is_lower
                token_flags["is_upper"] = token.is_upper
                token_flags["is_title"] = token.is_title
                token_flags["is_punct"] = token.is_punct
                token_flags["is_left_punct"] = token.is_left_punct
                token_flags["is_right_punct"] = token.is_right_punct
                token_flags["is_space"] = token.is_space
                token_flags["is_bracket"] = token.is_bracket
                token_flags["is_quote"] = token.is_quote
                token_flags["is_currency"] = token.is_currency
                token_flags["like_url"] = token.like_url
                token_flags["like_num"] = token.like_num
                token_flags["like_email"] = token.like_email
                token_flags["is_oov"] = token.is_oov
                token_flags["is_stop"] = token.is_stop
            if self.extract_sentiment:
                token_flags["sentiment"] = token.sentiment

            tokens.append(token.text)
            flags.append(token_flags)

        return tokens, flags



# coding: utf8

import bisect
import sys


def offset(id):
    return id[0] + str(int(id[1:]) + 1000)


class EntityAnnotation:
    def __init__(self, id, typ, spans, text):
        self.id = id
        self.type = typ
        self.spans = spans
        self.text = text

    @staticmethod
    def parse(line):
        id, mid, text = line.strip().split("\t")
        typ, spans = mid.split(" ", 1)
        spans = [tuple(s.split()) for s in spans.split(";")]
        return EntityAnnotation(id, typ, spans, text)

    def __repr__(self):
        return "<Entity(id=%r, type=%r, spans=%r, text=%r)>" % (
            self.id,
            self.type,
            self.spans,
            self.text,
        )

    def offset_id(self):
        self.id = offset(self.id)

    def as_brat(self):
        spans = ";".join(" ".join(s) for s in self.spans)
        return "%s\t%s %s\t%s" % (self.id, self.type, spans, self.text)


class RelationAnnotation:
    def __init__(self, id, typ, arg1, arg2):
        self.id = id
        self.type = typ
        self.arg1 = arg1
        self.arg2 = arg2

    @staticmethod
    def parse(line):
        id, typ, arg1, arg2 = line.strip().split()
        arg1 = arg1.split(":")[1]
        arg2 = arg2.split(":")[1]
        return RelationAnnotation(id, typ, arg1, arg2)

    def offset_id(self):
        self.arg1 = offset(self.arg1)
        self.arg2 = offset(self.arg2)
        self.id = offset(self.id)

    def __repr__(self):
        return "<Relation(id=%r, type=%r, arg1=%r, arg2=%r)>" % (
            self.id,
            self.type,
            self.arg1,
            self.arg2,
        )

    def as_brat(self):
        return "%s\t%s Arg1:%s Arg2:%s" % (self.id, self.type, self.arg1, self.arg2)


class SameAsAnnotation:
    total = 0

    def __init__(self, id, typ, args):
        self.id = id
        self.type = typ
        self.args = args

    @staticmethod
    def parse(line):
        SameAsAnnotation.total += 1
        typ, args = line[1:].strip().split(" ", 1)
        id = "*%d" % SameAsAnnotation.total
        args = args.split()
        return SameAsAnnotation(id, typ, args)

    def offset_id(self):
        self.args = [offset(arg) for arg in self.args]

    def __repr__(self):
        return "<Relation(id=%r, type=%r, args=%r)>" % (self.id, self.type, self.args)

    def as_brat(self):
        return "*\t%s %s" % (self.type, " ".join(self.args))


class EventAnnotation:
    def __init__(self, id, typ, ref, args):
        self.id = id
        self.type = typ
        self.ref = ref
        self.args = args

    @staticmethod
    def parse(line):
        id, mid = line.strip().split("\t")
        args = mid.split()
        typ, ref = args[0].split(":")
        args = args[1:]
        args = {arg.split(":")[0]: arg.split(":")[1] for arg in args}
        return EventAnnotation(id, typ, ref, args)

    def offset_id(self):
        self.ref = offset(self.ref)
        self.id = offset(self.id)

        for k in self.args:
            self.args[k] = offset(self.args[k])

    def __repr__(self):
        return "<Event(id=%r, type=%r, ref=%r, args=%r)>" % (
            self.id,
            self.type,
            self.ref,
            self.args,
        )

    def as_brat(self):
        spans = " ".join(k + ":" + v for k, v in self.args.items())
        return "%s\t%s:%s %s" % (self.id, self.type, self.ref, spans)


class AttributeAnnotation:
    def __init__(self, id, typ, ref):
        self.id = id
        self.type = typ
        self.ref = ref

    @staticmethod
    def parse(line):
        id, typ, ref = line.strip().split()
        return AttributeAnnotation(id, typ, ref)

    def offset_id(self):
        self.ref = offset(self.ref)
        self.id = offset(self.id)

    def __repr__(self):
        return "<Attribute(id=%r, type=%r, ref=%r)>" % (self.id, self.type, self.ref)

    def as_brat(self):
        return "%s\t%s %s" % (self.id, self.type, self.ref)


class AnnFile:
    def __init__(self):
        self.annotations = []

    def load(self, path):
        with open(path) as fp:
            for line in fp:
                ann = self._parse(line)
                if ann:
                    self.annotations.append(ann)

        return self

    def annotations_of(self, type):
        for e in self.annotations:
            if isinstance(e, type):
                yield e

    def filter_sentences(self, sentences, order):
        skip = 0
        sentence = 0
        free_space = 0

        skipped_space = []
        selected_sentence_spans = []
        while order:
            next_sentence = order.pop(0) - 1

            skip_backup = skip
            while sentence != next_sentence:
                skip += len(sentences.pop(0)) + 1
                sentence += 1

            free_space += skip - skip_backup

            current_length = len(sentences[0])
            selected_sentence_spans.append((skip, skip + current_length))
            skipped_space.append(free_space)
            free_space -= current_length + 1

        selected_annotations = {}
        for entity in self.annotations_of(EntityAnnotation):
            min_start = min(int(start) for start, _ in entity.spans)
            max_end = max(int(end) for _, end in entity.spans)
            try:
                sentence = next(
                    i
                    for i, (start, end) in enumerate(selected_sentence_spans)
                    if (start <= min_start and max_end <= end)
                )
            except StopIteration:
                continue

            entity.spans = [
                tuple(str(int(x) - skipped_space[sentence]) for x in span)
                for span in entity.spans
            ]
            selected_annotations[entity.id] = entity

        for ann in self.annotations_of(EventAnnotation):
            if ann.ref in selected_annotations:
                selected_annotations[ann.id] = ann

        for ann in self.annotations:
            add = (
                isinstance(ann, SameAsAnnotation)
                and ann.args[0] in selected_annotations
            )
            add |= (
                isinstance(ann, RelationAnnotation) and ann.arg1 in selected_annotations
            )
            add |= (
                isinstance(ann, AttributeAnnotation) and ann.ref in selected_annotations
            )
            if add:
                selected_annotations[ann.id] = ann

        self.annotations = list(selected_annotations.values())

    def offset_spans(self, sentences, first):
        sentences_offset = self._compute_sentence_offset(sentences)

        for ann in self.annotations_of(EntityAnnotation):
            locations = list(
                set(
                    [
                        bisect.bisect_left(sentences_offset, int(s))
                        for span in ann.spans
                        for s in span
                    ]
                )
            )

            if len(locations) != 1:
                raise ValueError()

            location = locations.pop()
            offset = sentences_offset[location] + 1

            if first:
                offset = sentences_offset[location - 1] + 1 if location > 0 else 0

            ann.spans = [
                (str(int(span[0]) + offset), str(int(span[1]) + offset))
                for span in ann.spans
            ]

    def _compute_sentence_offset(self, sentences):
        sentences_offset = [-1]

        for s in sentences:
            prev = sentences_offset[-1]
            start = prev + 1
            end = start + len(s)
            sentences_offset.append(end)

        sentences_offset.pop(0)
        return sentences_offset

    def offset_ids(self):
        for ann in self.annotations:
            ann.offset_id()

    def _parse(self, line):
        if line.startswith("T"):
            return EntityAnnotation.parse(line)

        if line.startswith("R"):
            return RelationAnnotation.parse(line)

        if line.startswith("*"):
            return SameAsAnnotation.parse(line)

        if line.startswith("E"):
            return EventAnnotation.parse(line)

        if line.startswith("A"):
            return AttributeAnnotation.parse(line)

        if line.startswith("#"):
            return None

        raise ValueError("Unknown annotation: %s" % line)


def merge(ann1: str, ann2: str, text: str):
    """Merge annotations of two different versions of the same file.
    """
    file1 = AnnFile().load(ann1)
    file2 = AnnFile().load(ann2)
    sents = open(text).read().split("\n")

    file1.offset_spans(sents, first=True)
    file2.offset_spans(sents, first=False)
    file2.offset_ids()

    for ann in file1.annotations:
        print(ann.as_brat())

    for ann in file2.annotations:
        print(ann.as_brat())


def review(ann: str, text: str, order: str):
    """Process a merged annotation file and outputs the selected annotations.
    """
    file1 = AnnFile().load(ann)
    sents = open(text).read().split("\n")
    order = open(order).read().split("\n")
    order = [int(line.strip("*")) for line in order if line]

    file1.filter_sentences(sents, order)
    for ann in file1.annotations:
        print(ann.as_brat())


def review_text(text: str, order: str):
    """Process a merged annotation file and outputs the selected sentences.
    """
    sents = open(text).read().split("\n")
    order = open(order).read().split("\n")
    order = [int(line.strip("*")) for line in order if line]

    selected = [sents[i - 1] for i in order]
    for sent in selected:
        print(sent)


def to_review(order: str):
    order = open(order).read().split("\n")
    for i, o in enumerate(order):
        if o.endswith("*"):
            print(i + 1)



import signal
import resource
import psutil
import warnings
import os
import multiprocessing


class ResourceManager:
    """
    Resource manager class.

    ##### Parameters

    - `time_limit: int`: maximum amount of seconds a function can run for (default = `5 minutes`).
    - `ram_limit: int`: maximum amount of memory bytes a function can use (default = `4 GB`).

    ##### Notes

    - Only one function may be restricted at the same time.
      Upcoming updates will fix this matter.
    - Memory restriction is issued upon the process's heap memory size and not
      over total address space in order to get a better estimation of the used memory.

    """

    def __init__(self, time_limit: int = 300, memory_limit: int = 4294967296):
        self.set_time_limit(time_limit)
        self.set_memory_limit(memory_limit)
        signal.signal(signal.SIGXCPU, alarm_handler)

    def set_memory_limit(self, limit):
        """
        Set the memory limit for future restricted functions.

        If memory limit is higher or equal than the current OS limit
        then it won't be changed.
        """
        soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
        self.original_limit = (soft, hard)
        self.memory_limit = None
        used_memory = self.get_used_memory()

        if limit <= (used_memory + 500 * 1024 ** 2):
            warnings.warn(
                "Especified memory limit is too close to the used amount. Will not be taken into account."
            )
            return

        if soft == -1 or limit < soft:
            self.memory_limit = (limit, hard)
        else:
            warnings.warn(
                "Especified memory limit is higher than OS limit. Will not be taken into account."
            )

    def set_time_limit(self, limit):
        self.time_limit = limit

    def _restrict(self, memory_amount):
        if memory_amount:
            _, hard = self.original_limit
            limit, _ = memory_amount
            resource.setrlimit(resource.RLIMIT_DATA, (limit, hard))
            resource.setrlimit(resource.RLIMIT_CPU, (self.time_limit, hard))

    def _unrestrict_memory(self):
        self._restrict(self.original_limit)

    def _run_for(self, function, *args, **kwargs):
        def signal_handler(*args):
            raise TimeoutError()

        try:
            signal.signal(signal.SIGALRM, signal_handler)
            signal.alarm(self.time_limit)

            result = function(*args, **kwargs)
            signal.alarm(0)  # cancel the alarm
            return result
        except Exception as e:
            # signal.alarm(0) #cancel the alarm
            raise e

    def get_used_memory(self):
        """
        Returns the amount of memory being used by the current process.
        """
        process = psutil.Process(os.getpid())
        return process.memory_info().rss

    def _restricted_function(self, result_bucket, function, args, kwargs):
        try:
            self._restrict(self.memory_limit)
            result = function(*args, **kwargs)
            result_bucket["result"] = result
        except Exception as e:
            result_bucket["result"] = e

    def run_restricted(self, function, *args, **kwargs):
        """
        Executes a given function with restricted amount of
        CPU time and RAM memory usage.
        """
        manager = multiprocessing.Manager()
        result_bucket = manager.dict()
        
        rprocess = multiprocessing.Process(target=self._restricted_function,
                                            args=[result_bucket, function, args, kwargs])

        rprocess.start()
        # print("started process:", rprocess.pid)
        rprocess.join()
        # print("ended process:", rprocess.pid)
        result = result_bucket["result"]
        if isinstance(result, Exception): #Exception ocurred
            raise result
        return result
        
        

def alarm_handler(*args):
    raise TimeoutError("process %d got to time limit" %os.getpid())



import enlighten
import warnings
import time
import datetime
import statistics
import math
import traceback
import sys
import termcolor

from autogoal.utils import ResourceManager, RestrictedWorkerByJoin, Min, Gb
from autogoal.sampling import ReplaySampler


class SearchAlgorithm:
    def __init__(
        self,
        generator_fn=None,
        fitness_fn=None,
        pop_size=1,
        maximize=True,
        errors="raise",
        early_stop=0.5,
        evaluation_timeout: int = 5 * Min,
        memory_limit: int = 4 * Gb,
        search_timeout: int = 60 * 60,
        target_fn=None,
        allow_duplicates=True,
    ):
        if generator_fn is None and fitness_fn is None:
            raise ValueError("You must provide either `generator_fn` or `fitness_fn`")

        self._generator_fn = generator_fn
        self._fitness_fn = fitness_fn or (lambda x: x)
        self._pop_size = pop_size
        self._maximize = maximize
        self._errors = errors
        self._evaluation_timeout = evaluation_timeout
        self._memory_limit = memory_limit
        self._early_stop = early_stop
        self._search_timeout = search_timeout
        self._target_fn = target_fn
        self._allow_duplicates = allow_duplicates

        if self._evaluation_timeout > 0 or self._memory_limit > 0:
            self._fitness_fn = RestrictedWorkerByJoin(
                self._fitness_fn, self._evaluation_timeout, self._memory_limit
            )

    def run(self, generations=None, logger=None):
        """Runs the search performing at most `generations` of `fitness_fn`.

        Returns:
            Tuple `(best, fn)` of the best found solution and its corresponding fitness.
        """
        if logger is None:
            logger = Logger()

        if generations is None:
            generations = math.inf

        if isinstance(logger, list):
            logger = MultiLogger(*logger)

        if isinstance(self._early_stop, float):
            early_stop = int(self._early_stop * generations)
        else:
            early_stop = self._early_stop

        best_solution = None
        best_fn = None
        no_improvement = 0
        start_time = time.time()
        seen = set()

        logger.begin(generations, self._pop_size)

        try:
            while generations > 0:
                stop = False

                logger.start_generation(generations, best_fn)
                self._start_generation()

                fns = []

                improvement = False

                for _ in range(self._pop_size):
                    solution = None

                    try:
                        solution = self._generate()
                    except Exception as e:
                        logger.error(
                            "Error while generating solution: %s" % e, solution
                        )
                        continue

                    if not self._allow_duplicates and repr(solution) in seen:
                        continue

                    try:
                        logger.sample_solution(solution)
                        fn = self._fitness_fn(solution)
                    except Exception as e:
                        fn = 0.0
                        logger.error(e, solution)

                        if self._errors == "raise":
                            logger.end(best_solution, best_fn)
                            raise e from None

                    if not self._allow_duplicates:
                        seen.add(repr(solution))

                    logger.eval_solution(solution, fn)
                    fns.append(fn)

                    if (
                        best_fn is None
                        or (fn > best_fn and self._maximize)
                        or (fn < best_fn and not self._maximize)
                    ):
                        logger.update_best(solution, fn, best_solution, best_fn)
                        best_solution = solution
                        best_fn = fn
                        improvement = True

                        if self._target_fn and best_fn >= self._target_fn:
                            stop = True
                            break

                    spent_time = time.time() - start_time

                    if self._search_timeout and spent_time > self._search_timeout:
                        print("(!) Stopping since time spent is %.2f." % (spent_time))
                        stop = True
                        break

                if not improvement:
                    no_improvement += 1
                else:
                    no_improvement = 0

                generations -= 1

                if generations <= 0:
                    print("(!) Stopping since all generations are done.")
                    stop = True
                    break

                if early_stop and no_improvement >= early_stop:
                    print(
                        "(!) Stopping since no improvement for %i generations."
                        % no_improvement
                    )
                    stop = True
                    break

                logger.finish_generation(fns)
                self._finish_generation(fns)

                if stop:
                    break

        except KeyboardInterrupt:
            pass

        logger.end(best_solution, best_fn)
        return best_solution, best_fn

    def _generate(self):
        # BUG: When multiprocessing is used for evaluation and no generation
        #      function is defined, the actual sampling occurs during fitness
        #      evaluation, and since that process has a copy of the solution
        #      we don't get the history in the `ReplaySampler`.

        sampler = ReplaySampler(self._build_sampler())

        if self._generator_fn is not None:
            solution = self._generator_fn(sampler)
        else:
            solution = sampler

        solution.sampler_ = sampler
        return solution

    def _build_sampler(self):
        raise NotImplementedError()

    def _start_generation(self):
        pass

    def _finish_generation(self, fns):
        pass


class Logger:
    def begin(self, generations, pop_size):
        pass

    def end(self, best, best_fn):
        pass

    def start_generation(self, generations, best_fn):
        pass

    def finish_generation(self, fns):
        pass

    def sample_solution(self, solution):
        pass

    def eval_solution(self, solution, fitness):
        pass

    def error(self, e: Exception, solution):
        pass

    def update_best(self, new_best, new_fn, previous_best, previous_fn):
        pass


class ConsoleLogger(Logger):
    def begin(self, generations, pop_size):
        print("Starting search: generations=%i" % generations)
        self.start_time = time.time()
        self.start_generations = generations

    @staticmethod
    def normal(text):
        return termcolor.colored(text, color="gray")

    @staticmethod
    def emph(text):
        return termcolor.colored(text, color="white", attrs=["bold"])

    @staticmethod
    def success(text):
        return termcolor.colored(text, color="green")

    @staticmethod
    def primary(text):
        return termcolor.colored(text, color="blue")

    @staticmethod
    def warn(text):
        return termcolor.colored(text, color="orange")

    @staticmethod
    def err(text):
        return termcolor.colored(text, color="red")

    def start_generation(self, generations, best_fn):
        current_time = time.time()
        elapsed = int(current_time - self.start_time)
        avg_time = elapsed / (self.start_generations - generations + 1)
        remaining = int(avg_time * generations)
        elapsed = datetime.timedelta(seconds=elapsed)
        remaining = datetime.timedelta(seconds=remaining)

        print(
            self.emph("New generation started"),
            self.success(f"best_fn={float(best_fn or 0.0):0.3}"),
            self.primary(f"generations={generations}"),
            self.primary(f"elapsed={elapsed}"),
            self.primary(f"remaining={remaining}"),
        )

    def error(self, e: Exception, solution):
        print(self.err("(!) Error evaluating pipeline: %s" % e))

    def end(self, best, best_fn):
        print(self.emph("Search completed: best_fn=%.3f, best=\n%r" % (best_fn, best)))

    def sample_solution(self, solution):
        print(self.emph("Evaluating pipeline:"))
        print(solution)

    def eval_solution(self, solution, fitness):
        print(self.primary("Fitness=%.3f" % fitness))

    def update_best(self, new_best, new_fn, previous_best, previous_fn):
        print(
            self.success(
                "Best solution: improved=%.3f, previous=%.3f"
                % (new_fn, previous_fn or 0)
            )
        )


class ProgressLogger(Logger):
    def begin(self, generations, pop_size):
        self.manager = enlighten.get_manager()
        self.pop_counter = self.manager.counter(
            total=pop_size, unit="evals", leave=True, desc="Current Gen"
        )
        self.total_counter = self.manager.counter(
            total=generations * pop_size, unit="evals", leave=True, desc="Best: 0.000"
        )

    def sample_solution(self, solution):
        self.pop_counter.update()
        self.total_counter.update()

    def start_generation(self, generations, best_fn):
        self.pop_counter.count = 0
        self.total_counter.update(force=True)

    def update_best(self, new_best, new_fn, *args):
        self.total_counter.desc = "Best: %.3f" % new_fn

    def end(self, *args):
        self.pop_counter.close()
        self.total_counter.close()
        self.manager.stop()


class MemoryLogger(Logger):
    def __init__(self):
        self.generation_best_fn = [0]
        self.generation_mean_fn = []

    def update_best(self, new_best, new_fn, previous_best, previous_fn):
        self.generation_best_fn[-1] = new_fn

    def finish_generation(self, fns):
        try:
            mean = statistics.mean(fns)
        except:
            mean = 0
        self.generation_mean_fn.append(mean)
        self.generation_best_fn.append(self.generation_best_fn[-1])


class MultiLogger(Logger):
    def __init__(self, *loggers):
        self.loggers = loggers

    def run(self, name, *args, **kwargs):
        for logger in self.loggers:
            getattr(logger, name)(*args, **kwargs)

    def begin(self, *args, **kwargs):
        self.run("begin", *args, **kwargs)

    def end(self, *args, **kwargs):
        self.run("end", *args, **kwargs)

    def start_generation(self, *args, **kwargs):
        self.run("start_generation", *args, **kwargs)

    def finish_generation(self, *args, **kwargs):
        self.run("finish_generation", *args, **kwargs)

    def sample_solution(self, *args, **kwargs):
        self.run("sample_solution", *args, **kwargs)

    def eval_solution(self, *args, **kwargs):
        self.run("eval_solution", *args, **kwargs)

    def error(self, *args, **kwargs):
        self.run("error", *args, **kwargs)

    def update_best(self, *args, **kwargs):
        self.run("update_best", *args, **kwargs)



from transformers import BertModel, BertTokenizer
from pathlib import Path
import torch
import numpy as np

from autogoal.kb import Sentence, MatrixContinuousDense, Tensor3, List, Word
from autogoal.grammar import Discrete, Categorical
from autogoal.utils import CacheManager, nice_repr


@nice_repr
class BertEmbedding:
    """
    Transforms a sentence already tokenized into a list of vector embeddings using a Bert pretrained multilingual model.

    ##### Examples

    ```python
    >>> sentence = "embed this wrongword".split()
    >>> bert = BertEmbedding(verbose=False)
    >>> embedding = bert.run(sentence)
    >>> embedding.shape
    (3, 768)
    >>> embedding
    array([[ 0.3887945 , -0.22509816,  0.24768752, ...,  0.7490128 ,
             0.00565467, -0.2144883 ],
           [ 0.1428812 , -0.25218996,  0.19961214, ...,  0.964931  ,
             0.5816741 , -0.2297722 ],
           [ 0.63840234, -0.09097156, -0.80802155, ...,  0.9195696 ,
             0.27364567,  0.14955777]], dtype=float32)

    ```

    ##### Notes

    On the first use the model `best-base-multilingual-cased` 
    from [huggingface/transformers](https://github.com/huggingface/transformers)
    will be downloaded. This may take a few minutes.

    If you are using the development container the model should be already downloaded for you.
    """
    use_cache = False

    def __init__(self, merge_mode: Categorical("avg", "first") = "avg", *, verbose=False):  # , length: Discrete(16, 512)):
        self.device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )
        self.verbose = verbose
        self.print("Using device: %s" % self.device)
        self.merge_mode = merge_mode
        self.model = None
        self.tokenizer = None

    def print(self, *args, **kwargs):
        if not self.verbose:
            return

        print(*args, **kwargs)

    def run(self, input: List(Word(language="english"))) -> MatrixContinuousDense():
        if self.model is None:
            self.model = BertModel.from_pretrained("bert-base-multilingual-cased").to(self.device)
            self.tokenizer = BertTokenizer.from_pretrained("bert-base-multilingual-cased")

        if self.use_cache:
            sentence_hash = hash(" ".join(input))
            cache_path = Path(__file__).parent / f"cached_bert_{sentence_hash}.npy"

            if cache_path.exists():
                return np.load(cache_path)

        bert_tokens = [self.tokenizer.tokenize(x) for x in input]
        bert_senquence = self.tokenizer.encode_plus(
            [t for tokens in bert_tokens for t in tokens], return_tensors="pt"
        )
        with torch.no_grad():
            encoded_tokens, _ = self.model(**bert_senquence)
            encoded_tokens = encoded_tokens.squeeze(0)

        count = 0
        matrix = []
        for i, token in enumerate(input):
            contiguous = len(bert_tokens[i])
            vectors = encoded_tokens[count : count + contiguous, :]
            vector = self._merge(vectors)
            matrix.append(vector)
            count += contiguous

        matrix = np.vstack(matrix)

        if self.use_cache:
            np.save(cache_path, matrix)

        return matrix

    def _merge(self, vectors):
        if not vectors.size(0):
            return np.zeros(vectors.size(1), dtype="float32")
        if self.merge_mode == "avg":
            return vectors.mean(dim=0).numpy()
        elif self.merge_mode == "first":
            return vectors[0, :].numpy()
        else:
            raise ValueError("Unknown merge mode")


@nice_repr
class BertTokenizeEmbedding:
    """
    Transforms a sentence into a list of vector embeddings using a Bert pretrained English model.

    ##### Notes

    On the first use the model `best-base-multilingual-cased` from 
    [huggingface/transformers](https://github.com/huggingface/transformers)
    will be downloaded. This may take a few minutes.

    If you are using the development container the model should be already downloaded for you.
    """

    def __init__(self, verbose=False):  # , length: Discrete(16, 512)):
        self.device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )
        self.verbose = verbose
        self.print("Using device: %s" % self.device)
        self.model = None
        self.tokenizer = None

    def print(self, *args, **kwargs):
        if not self.verbose:
            return

        print(*args, **kwargs)

    def run(self, input: List(Sentence(language="english"))) -> Tensor3():
        if self.model is None:
            self.model = BertModel.from_pretrained("bert-base-multilingual-cased").to(self.device)
            self.tokenizer = BertTokenizer.from_pretrained("bert-base-multilingual-cased")

        self.print("Tokenizing...", end="", flush=True)
        tokens = [
            self.tokenizer.encode(x, max_length=32, pad_to_max_length=True)
            for x in input
        ]
        self.print("done")

        ids = torch.tensor(tokens).to(self.device)

        with torch.no_grad():
            self.print("Embedding...", end="", flush=True)
            output = self.model(ids)[0].numpy()
            self.print("done")

        return output



from ._base import Grammar, Sampler
from ._cfg import (
    generate_cfg,
    ContextFreeGrammar,
    Discrete,
    Continuous,
    Categorical,
    Boolean,
    Union,
    Symbol,
    CfgInitializer,
    Empty,
    Symbol,
    Subset,
)
from ._graph import GraphGrammar, Path, Block, Graph, GraphSpace, Epsilon



try:
    import nltk
except:
    print("(!) Code in `autogoal.contrib.nltk` requires `nltk`.")
    print("(!) You can install it with `pip install autogoal[nltk]`.")
    raise

from autogoal.contrib.nltk._generated import *



import multiprocessing
import warnings
import resource
import psutil
import signal
import os
import sys
import traceback

from autogoal.utils import Mb


class RestrictedWorker:
    def __init__(self, function, timeout: int, memory: int):
        self.function = function
        self.timeout = timeout
        self.memory = memory
        signal.signal(signal.SIGXCPU, alarm_handler)

    def _restrict(self):
        msoft, mhard = resource.getrlimit(resource.RLIMIT_DATA)
        csoft, chard = resource.getrlimit(resource.RLIMIT_CPU)
        used_memory = self.get_used_memory()

        if self.memory and self.memory > (used_memory + 500 * Mb):
            # memory may be restricted
            resource.setrlimit(resource.RLIMIT_DATA, (self.memory, mhard))
        else:
            warnings.warn("Cannot restrict memory")

        if self.timeout:
            # time may be restricted
            resource.setrlimit(resource.RLIMIT_CPU, (self.timeout, chard))
        else:
            warnings.warn("Cannot restrict cpu time")

    def _restricted_function(self, result_bucket, args, kwargs):
        try:
            self._restrict()
            result = self.function(*args, **kwargs)
            result_bucket["result"] = result
        except Exception as e:
            msg = "{}\n\nOriginal {}".format(e, traceback.format_exc())
            result_bucket["result"] = e.__class__(msg)

    def run_restricted(self, *args, **kwargs):
        """
        Executes a given function with restricted amount of
        CPU time and RAM memory usage
        """
        manager = multiprocessing.Manager()
        result_bucket = manager.dict()

        rprocess = multiprocessing.Process(
            target=self._restricted_function, args=[result_bucket, args, kwargs]
        )

        rprocess.start()
        rprocess.join()
        result = result_bucket["result"]

        if isinstance(result, Exception):  # Exception ocurred
            raise result

        return result

    def get_used_memory(self):
        """
        returns the amount of memory being used by the current process
        """
        process = psutil.Process(os.getpid())
        return process.memory_info().rss

    def __call__(self, *args, **kwargs):
        return self.run_restricted(*args, **kwargs)


def alarm_handler(*args):
    raise TimeoutError("process %d got to time limit" % os.getpid())


class RestrictedWorkerByJoin(RestrictedWorker):
    def __init__(self, function, timeout: int, memory: int):
        self.function = function
        self.timeout = timeout
        self.memory = memory

    def _restrict(self):
        _, mhard = resource.getrlimit(resource.RLIMIT_AS)
        used_memory = self.get_used_memory()

        if self.memory is None:
            return

        if self.memory > (used_memory + 50 * Mb):
            # memory may be restricted
            print("Restricting memory to %s" % self.memory, flush=True)
            resource.setrlimit(resource.RLIMIT_DATA, (self.memory, mhard))
        else:
            raise ValueError("Cannot restrict memory to %s < %i" % (self.memory, used_memory + 50 * Mb))

    def run_restricted(self, *args, **kwargs):
        """
        Executes a given function with restricted amount of
        CPU time and RAM memory usage
        """
        manager = multiprocessing.Manager()
        result_bucket = manager.dict()

        rprocess = multiprocessing.Process(
            target=self._restricted_function, args=[result_bucket, args, kwargs]
        )

        rprocess.start()
        rprocess.join(self.timeout)

        if rprocess.exitcode == 0:
            result = result_bucket["result"]
        else:
            rprocess.terminate()
            raise TimeoutError()

        if isinstance(result, Exception):  # Exception ocurred
            raise result

        return result



import time
import textwrap

from autogoal.search import Logger
from telegram.ext import Updater, Dispatcher, CommandHandler
from telegram import ParseMode


class TelegramLogger(Logger):
    def __init__(self, token, channel:str=None, name=""):
        self.name = name
        self.channel = int(channel) if channel and channel.isdigit() else channel
        self.last_time = time.time()
        self.updater = Updater(token)
        self.dispatcher = self.updater.dispatcher
        self.progress = 0
        self.generations = 1
        self.best = 0.0
        self.current = ""
        self.message = self.message = self.dispatcher.bot.send_message(
            chat_id=self.channel, text=f"**{self.name}**\nStarting...", parse_mode=ParseMode.MARKDOWN,
        )

    def begin(self, generations, pop_size):
        self.generations = generations
        self._send()

    def update_best(self, new_best, new_fn, *args):
        self.best = new_fn
        self._send()

    def end(self, best, best_fn):
        self.best = best_fn
        self._send()

    def eval_solution(self, solution, fitness):
        self.progress += 1
        self._send()

    def _send(self):
        if not self.channel:
            return

        if time.time() - self.last_time < 10:
            return

        self.last_time = time.time()

        text = textwrap.dedent(
            f"""
            **{self.name}**
            Best: `{float(self.best):0.3}`
            Iterations: `{self.progress}/{self.generations}`
            """
        )
        try:
            self.message.edit_text(
                text=text, parse_mode=ParseMode.MARKDOWN,
            )
        except:
            pass



from autogoal.kb import (
    algorithm,
    Sentence,
    List,
    Word,
    ContinuousVector,
    MatrixContinuousDense,
    Tensor3,
    Flags,
    Document,
    Distinct,
    Entity,
    Category,
    Tuple,
    Postag,
    Vector,
    CategoricalVector,
)
from autogoal.grammar import Categorical, Boolean
from autogoal.utils import nice_repr

import numpy as np


@nice_repr
class VectorAggregator:
    def __init__(self, mode: Categorical("mean", "max")):
        self.mode = mode

    def run(self, input: List(ContinuousVector())) -> ContinuousVector():
        input = np.vstack(input)

        if self.mode == "mean":
            return input.mean(axis=1)
        elif self.mode == "max":
            return input.max(axis=1)

        raise ValueError("Invalid mode: %s" % self.mode)


@nice_repr
class MatrixBuilder:
    """
    Builds a matrix from a list of vectors.

    ##### Examples

    ```python
    >>> import numpy as np
    >>> x1 = np.asarray([1,2,3])
    >>> x2 = np.asarray([2,3,4])
    >>> x3 = np.asarray([3,4,5])
    >>> MatrixBuilder().run([x1, x2, x3])
    array([[1, 2, 3],
           [2, 3, 4],
           [3, 4, 5]])

    ```
    """

    def run(self, input: List(ContinuousVector())) -> MatrixContinuousDense():
        return np.vstack(input)


@nice_repr
class TensorBuilder:
    """
    Builds a 3D tensor from a list of matrices.

    ##### Examples

    ```python
    >>> import numpy as np
    >>> x1 = np.asarray([[1,2],[3,4]])
    >>> x2 = np.asarray([[2,3],[4,5]])
    >>> x3 = np.asarray([[3,4],[5,6]])
    >>> TensorBuilder().run([x1, x2, x3])
    array([[[1, 2],
            [3, 4]],
    <BLANKLINE>
           [[2, 3],
            [4, 5]],
    <BLANKLINE>
           [[3, 4],
            [5, 6]]])

    ```
    """

    def run(self, input: List(MatrixContinuousDense())) -> Tensor3():
        return np.vstack([np.expand_dims(m, axis=0) for m in input])


@nice_repr
class FlagsMerger:
    def run(self, input: List(Flags())) -> Flags():
        result = {}

        for d in input:
            result.update(d)

        return result


@nice_repr
class MultipleFeatureExtractor:
    def __init__(
        self,
        extractors: Distinct(
            algorithm(Word(), Flags()), exceptions=["MultipleFeatureExtractor"]
        ),
        merger: algorithm(List(Flags()), Flags()),
    ):
        self.extractors = extractors
        self.merger = merger

    def run(self, input: Word()) -> Flags():
        flags = [extractor.run(input) for extractor in self.extractors]
        return self.merger.run(flags)


@nice_repr
class SentenceFeatureExtractor:
    def __init__(
        self,
        tokenizer: algorithm(Sentence(), List(Word())),
        feature_extractor: algorithm(Word(), Flags()),
        include_text: Boolean(),
    ):
        self.tokenizer = tokenizer
        self.feature_extractor = feature_extractor
        self.include_text = include_text

    def run(self, input: Sentence()) -> Flags():
        tokens = self.tokenizer.run(input)
        flags = [self.feature_extractor(w) for w in tokens]

        if self.include_text:
            return {
                f"{w}|{f}": v for w, flag in zip(tokens, flags) for f, v in flag.items()
            }
        else:
            return {f: v for flag in flags for f, v in flag.items()}


@nice_repr
class DocumentFeatureExtractor:
    def __init__(
        self,
        tokenizer: algorithm(Document(), List(Sentence())),
        feature_extractor: algorithm(Sentence(), Flags()),
    ):
        self.tokenizer = tokenizer
        self.feature_extractor = feature_extractor

    def run(self, input: Document()) -> List(Flags()):
        tokens = self.tokenizer.run(input)
        flags = [self.feature_extractor(w) for w in tokens]
        return


@nice_repr
class TextEntityEncoder:
    """
    Convierte una oración en texto plano más la lista de entidades
    reconocidas en la oración, en una lista de tokens con sus respectivas
    categorias BILOUV.
    """

    def __init__(self, tokenizer: algorithm(Sentence(), List(Word()))):
        self.tokenizer = tokenizer

    def run(
        self, input: Tuple(Sentence(), List(Entity()))
    ) -> Tuple(List(Word()), List(Postag())):
        pass


@nice_repr
class TextRelationEncoder:
    """
    Convierte una oración en texto plano y una lista de relaciones 
    que se cumplen entre entidades, en una lista de ejemplos
    por cada oración.
    """

    def __init__(self,
        tokenizer: algorithm(Sentence(), List(Word())),
        token_feature_extractor: algorithm(Word(), Flags()),
        # token_sentence_encoder: algorithm(Word(), )
    ):
        pass

    def run(
        self, input: Tuple(Sentence(), List(Tuple(Entity(), Entity(), Category())))
    ) -> Tuple(List(Vector()), CategoricalVector()):
        pass



try:
    import torch

    # assert sklearn.__version__ == "0.22"
except:
    print("(!) Code in `autogoal.contrib.torch` requires `pytorch==0.1.4`.")
    print("(!) You can install it with `pip install autogoal[torch]`.")
    raise


from ._bert import BertEmbedding, BertTokenizeEmbedding



try:
    import gensim

    # assert gensim.__version__ == "3.8.1"
except:
    print("(!) Code in `autogoal.contrib.gensim` requires `gensim==3.8.1`.")
    print("(!) You can install it with `pip install autogoal[gensim]`.")
    raise


from ._base import (
    Word2VecEmbedding, 
    Word2VecEmbeddingSpanish, 
    FastTextEmbeddingSpanishSUC, 
    FastTextEmbeddingSpanishSWBC, 
    GloveEmbeddingSpanishSWBC, 
)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Orginal implementation from keras_contrib/layers/crf
# ==============================================================================
"""Implementing Conditional Random Field layer."""

# !!! warning 
# This code is taken verbatim from <https://github.com/tensorflow/addons/pull/377>
# since the CRF layer has not yet landed in tensorflow
# TODO: Make sure to replace this when tensorflow merges this commit

# Files:
# <https://github.com/tensorflow/addons/blob/e0b5e0f056d24b95546970acd4aedc3a5530d466/tensorflow_addons/layers/crf.py>
# <https://github.com/tensorflow/addons/blob/e0b5e0f056d24b95546970acd4aedc3a5530d466/tensorflow_addons/losses/crf.py>

from __future__ import absolute_import, division, print_function

import tensorflow as tf
from typeguard import typechecked

from tensorflow_addons.text.crf import crf_decode, crf_log_likelihood
from tensorflow_addons.utils import types


class CRF(tf.keras.layers.Layer):
    """Linear chain conditional random field (CRF).

    Examples:

    ```python
        from tensorflow_addons.layers import CRF
        from tensorflow_addons.losses import crf_loss

        model = Sequential()
        model.add(Embedding(3001, 300, mask_zero=True)

        crf = CRF(10)
        model.add(crf)

        model.compile('adam', loss=crf_loss)

        model.fit(x, y)
    ```

    Arguments:
        units: Positive integer, dimensionality of the output space,
            should equal to tag num.
        chain_initializer: Initializer for the `chain_kernel` weights matrix,
            used for the CRF chain energy.
            (see [initializers](../initializers.md)).
        chain_regularizer: Regularizer function applied to
            the `chain_kernel` weights matrix.
        chain_constraint: Constraint function applied to
            the `chain_kernel` weights matrix.
        use_boundary: Boolean (default True), indicating if trainable
            start-end chain energies should be added to model.
        boundary_initializer: Initializer for the `left_boundary`,
            'right_boundary' weights vectors,
            used for the start/left and end/right boundary energy.
        boundary_regularizer: Regularizer function applied to
            the 'left_boundary', 'right_boundary' weight vectors.
        boundary_constraint: Constraint function applied to
            the `left_boundary`, `right_boundary` weights vectors.
        use_kernel: Boolean (default True), indicating if apply
            a fully connected layer before CRF op.
        kernel_initializer: Initializer for the `kernel` weights matrix,
            used for the linear transformation of the inputs.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
        use_bias: Boolean (default True), whether the layer uses a bias vector.
        bias_initializer: Initializer for the bias vector.
        bias_regularizer: Regularizer function applied to the bias vector.
        bias_constraint: Constraint function applied to the bias vector.
        activation: default value is 'linear', Activation function to use.

    Input shape:
        3D tensor with shape: `(batch_size, sequence_length, feature_size)`.

    Output shape:
        2D tensor (dtype: int32) with shape: `(batch_size, sequence_length)`.

    Masking:
        This layer supports masking
        (2D tensor, shape: `(batch_size, sequence_length)`)
        for input data with a variable number of timesteps.
        This layer output same make tensor,
        NOTICE this may cause issue when you
        use some keras loss and metrics function which usually expect 1D mask.

    Loss function:
        Due to the TF 2.0 version support eager execution be default,
        there is no way can implement CRF loss as independent loss function.
        Thus, user should use loss method of this layer.
        See Examples (above) for detailed usage.

    References:
        - [Conditional Random Field](https://en.wikipedia.org/wiki/Conditional_random_field)
    """

    @typechecked
    def __init__(
        self,
        units: int,
        chain_initializer: types.Initializer = "orthogonal",
        chain_regularizer: types.Regularizer = None,
        chain_constraint: types.Constraint = None,
        use_boundary: bool = True,
        boundary_initializer: types.Initializer = "zeros",
        boundary_regularizer: types.Regularizer = None,
        boundary_constraint: types.Constraint = None,
        use_kernel: bool = True,
        kernel_initializer: types.Initializer = "glorot_uniform",
        kernel_regularizer: types.Regularizer = None,
        kernel_constraint: types.Constraint = None,
        use_bias: bool = True,
        bias_initializer: types.Initializer = "zeros",
        bias_regularizer: types.Regularizer = None,
        bias_constraint: types.Constraint = None,
        activation: types.Activation = "linear",
        **kwargs
    ):
        super(CRF, self).__init__(**kwargs)

        # setup mask supporting flag, used by base class (the Layer)
        # because base class's init method will set it to False unconditionally
        # So this assigned must be executed after call base class's init method
        self.supports_masking = True

        self.units = units  # numbers of tags

        self.use_boundary = use_boundary
        self.use_bias = use_bias
        self.use_kernel = use_kernel

        self.activation = tf.keras.activations.get(activation)

        self.kernel_initializer = tf.keras.initializers.get(kernel_initializer)
        self.chain_initializer = tf.keras.initializers.get(chain_initializer)
        self.boundary_initializer = tf.keras.initializers.get(boundary_initializer)
        self.bias_initializer = tf.keras.initializers.get(bias_initializer)

        self.kernel_regularizer = tf.keras.regularizers.get(kernel_regularizer)
        self.chain_regularizer = tf.keras.regularizers.get(chain_regularizer)
        self.boundary_regularizer = tf.keras.regularizers.get(boundary_regularizer)
        self.bias_regularizer = tf.keras.regularizers.get(bias_regularizer)

        self.kernel_constraint = tf.keras.constraints.get(kernel_constraint)
        self.chain_constraint = tf.keras.constraints.get(chain_constraint)
        self.boundary_constraint = tf.keras.constraints.get(boundary_constraint)
        self.bias_constraint = tf.keras.constraints.get(bias_constraint)

        # values will be assigned in method
        self.input_spec = None

        # value remembered for loss/metrics function
        self.potentials = None
        self.sequence_length = None
        self.mask = None

        # global variable
        self.chain_kernel = None
        self._dense_layer = None
        self.left_boundary = None
        self.right_boundary = None

    def build(self, input_shape):
        input_shape = tuple(tf.TensorShape(input_shape).as_list())

        # see API docs of InputSpec for more detail
        self.input_spec = [tf.keras.layers.InputSpec(shape=input_shape)]

        # weights that work as transfer probability of each tags
        self.chain_kernel = self.add_weight(
            shape=(self.units, self.units),
            name="chain_kernel",
            initializer=self.chain_initializer,
            regularizer=self.chain_regularizer,
            constraint=self.chain_constraint,
        )

        # weight of <START> to tag probability and tag to <END> probability
        if self.use_boundary:
            self.left_boundary = self.add_weight(
                shape=(self.units,),
                name="left_boundary",
                initializer=self.boundary_initializer,
                regularizer=self.boundary_regularizer,
                constraint=self.boundary_constraint,
            )
            self.right_boundary = self.add_weight(
                shape=(self.units,),
                name="right_boundary",
                initializer=self.boundary_initializer,
                regularizer=self.boundary_regularizer,
                constraint=self.boundary_constraint,
            )

        if self.use_kernel:
            self._dense_layer = tf.keras.layers.Dense(
                units=self.units,
                activation=self.activation,
                use_bias=self.use_bias,
                bias_initializer=self.bias_initializer,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                kernel_constraint=self.kernel_constraint,
                bias_constraint=self.bias_constraint,
                dtype=self.dtype,
            )
        else:
            self._dense_layer = lambda x: tf.cast(x, dtype=self.dtype)

        super(CRF, self).build(input_shape)

    def call(self, inputs, mask=None, **kwargs):
        # mask: Tensor(shape=(batch_size, sequence_length), dtype=bool) or None

        if mask is not None:
            if tf.keras.backend.ndim(mask) != 2:
                raise ValueError("Input mask to CRF must have dim 2 if not None")

        # left padding of mask is not supported, due the underline CRF function
        # detect it and report it to user
        first_mask = None
        if mask is not None:
            left_boundary_mask = self._compute_mask_left_boundary(mask)
            first_mask = left_boundary_mask[:, 0]

        # remember this value for later use
        self.mask = mask

        if first_mask is not None:
            no_left_padding = tf.math.reduce_all(first_mask)
            msg = "Currently, CRF layer do not support left padding"
            with tf.control_dependencies(
                [
                    tf.debugging.assert_equal(
                        no_left_padding, tf.constant(True), message=msg
                    )
                ]
            ):
                self.potentials = self._dense_layer(inputs)
        else:
            self.potentials = self._dense_layer(inputs)

        # appending boundary probability info
        if self.use_boundary:
            self.potentials = self.add_boundary_energy(
                self.potentials, mask, self.left_boundary, self.right_boundary
            )

        self.sequence_length = self._get_sequence_length(inputs, mask)

        decoded_sequence, _ = self.get_viterbi_decoding(
            self.potentials, self.sequence_length
        )

        return decoded_sequence

    def _get_sequence_length(self, input_, mask):
        """Currently underline CRF fucntion (provided by
        tensorflow_addons.text.crf) do not support bi-direction masking (left
        padding / right padding), it support right padding by tell it the
        sequence length.

        this function is compute the sequence length from input and
        mask.
        """
        if mask is not None:
            sequence_length = self.mask_to_sequence_length(mask)
        else:
            # make a mask tensor from input, then used to generate sequence_length
            input_energy_shape = tf.shape(input_)
            raw_input_shape = tf.slice(input_energy_shape, [0], [2])
            alt_mask = tf.ones(raw_input_shape)

            sequence_length = self.mask_to_sequence_length(alt_mask)

        return sequence_length

    def mask_to_sequence_length(self, mask):
        """compute sequence length from mask."""
        sequence_length = tf.cast(tf.reduce_sum(tf.cast(mask, tf.int8), 1), tf.int64)
        return sequence_length

    @staticmethod
    def _compute_mask_right_boundary(mask):
        """input mask: 0011100, output left_boundary: 0000100."""
        # shift mask to left by 1: 0011100 => 0111000
        offset = 1
        left_shifted_mask = tf.concat(
            [mask[:, offset:], tf.zeros_like(mask[:, :offset])], axis=1
        )

        # NOTE: below code is different from keras_contrib
        # Original code in keras_contrib:
        # end_mask = K.cast(
        #   K.greater(self.shift_left(mask), mask),
        #   K.floatx()
        # )
        # has a bug, confirmed
        # by the original keras_contrib maintainer
        # Luiz Felix (github: lzfelix),

        # 0011100 > 0111000 => 0000100
        right_boundary = tf.greater(mask, left_shifted_mask)

        return right_boundary

    @staticmethod
    def _compute_mask_left_boundary(mask):
        """input mask: 0011100, output left_boundary: 0010000."""
        # shift mask to right by 1: 0011100 => 0001110
        offset = 1
        right_shifted_mask = tf.concat(
            [tf.zeros_like(mask[:, :offset]), mask[:, :-offset]], axis=1
        )

        # 0011100 > 0001110 => 0010000
        left_boundary = tf.greater(
            tf.cast(mask, tf.int32), tf.cast(right_shifted_mask, tf.int32)
        )
        # left_boundary = tf.greater(mask, right_shifted_mask)

        return left_boundary

    def add_boundary_energy(self, potentials, mask, start, end):
        def expand_scalar_to_3d(x):
            # expand tensor from shape (x, ) to (1, 1, x)
            return tf.reshape(x, (1, 1, -1))

        start = expand_scalar_to_3d(start)
        end = expand_scalar_to_3d(end)
        if mask is None:
            potentials = tf.concat(
                [potentials[:, :1, :] + start, potentials[:, 1:, :]], axis=1
            )
            potentials = tf.concat(
                [potentials[:, :-1, :], potentials[:, -1:, :] + end], axis=1
            )
        else:
            mask = tf.keras.backend.expand_dims(tf.cast(mask, start.dtype), axis=-1)
            start_mask = tf.cast(self._compute_mask_left_boundary(mask), start.dtype)

            end_mask = tf.cast(self._compute_mask_right_boundary(mask), end.dtype)
            potentials = potentials + start_mask * start
            potentials = potentials + end_mask * end
        return potentials

    def get_viterbi_decoding(self, potentials, sequence_length):
        # decode_tags: A [batch_size, max_seq_len] matrix, with dtype `tf.int32`
        decode_tags, best_score = crf_decode(
            potentials, self.chain_kernel, sequence_length
        )

        return decode_tags, best_score

    def get_config(self):
        # used for loading model from disk
        config = {
            "units": self.units,
            "use_boundary": self.use_boundary,
            "use_bias": self.use_bias,
            "use_kernel": self.use_kernel,
            "kernel_initializer": tf.keras.initializers.serialize(
                self.kernel_initializer
            ),
            "chain_initializer": tf.keras.initializers.serialize(
                self.chain_initializer
            ),
            "boundary_initializer": tf.keras.initializers.serialize(
                self.boundary_initializer
            ),
            "bias_initializer": tf.keras.initializers.serialize(self.bias_initializer),
            "activation": tf.keras.activations.serialize(self.activation),
            "kernel_regularizer": tf.keras.regularizers.serialize(
                self.kernel_regularizer
            ),
            "chain_regularizer": tf.keras.regularizers.serialize(
                self.chain_regularizer
            ),
            "boundary_regularizer": tf.keras.regularizers.serialize(
                self.boundary_regularizer
            ),
            "bias_regularizer": tf.keras.regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": tf.keras.constraints.serialize(self.kernel_constraint),
            "chain_constraint": tf.keras.constraints.serialize(self.chain_constraint),
            "boundary_constraint": tf.keras.constraints.serialize(
                self.boundary_constraint
            ),
            "bias_constraint": tf.keras.constraints.serialize(self.bias_constraint),
        }
        base_config = super(CRF, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        output_shape = input_shape[:2]
        return output_shape

    def compute_mask(self, input_, mask=None):
        """keep mask shape [batch_size, max_seq_len]"""
        return mask

    def get_negative_log_likelihood(self, y_true):
        y_true = tf.cast(y_true, tf.int32)
        self.sequence_length = tf.cast(self.sequence_length, tf.int32)

        log_likelihood, _ = crf_log_likelihood(
            self.potentials, y_true, self.sequence_length, self.chain_kernel
        )

        return -log_likelihood

    def get_loss(self, y_true, y_pred):
        # we don't use y_pred, but caller pass it anyway, ignore it
        return self.get_negative_log_likelihood(y_true)

    def get_accuracy(self, y_true, y_pred):
        judge = tf.cast(tf.equal(y_pred, y_true), tf.keras.backend.floatx())
        if self.mask is None:
            return tf.reduce_mean(judge)
        else:
            mask = tf.cast(self.mask, tf.keras.backend.floatx())
            return tf.reduce_sum(judge * mask) / tf.reduce_sum(mask)

    def __call__(self, inputs, *args, **kwargs):
        outputs = super(CRF, self).__call__(inputs, *args, **kwargs)

        # A hack that add _keras_history to EagerTensor, make it more like normal Tensor
        for tensor in tf.nest.flatten(outputs):
            if not hasattr(tensor, "_keras_history"):
                tensor._keras_history = (self, 0, 0)

        return outputs

    @property
    def _compute_dtype(self):
        # fixed output dtype from underline CRF functions
        return tf.int32


class ConditionalRandomFieldLoss(object):
    def __init__(self, name: str = "crf_loss"):
        self.name = name

    def get_config(self):
        return {"name": self.name}

    def __call__(self, y_true, y_pred, sample_weight=None):
        crf_layer = y_pred._keras_history[0]

        # check if last layer is CRF
        if not isinstance(crf_layer, CRF):
            raise ValueError(
                "Last layer must be CRF for use {}.".format(self.__class__.__name__)
            )

        loss_vector = crf_layer.get_loss(y_true, y_pred)

        return tf.keras.backend.mean(loss_vector)


crf_loss = ConditionalRandomFieldLoss()



import re
import abc

from autogoal.utils import nice_repr
from autogoal.kb import Word, Flags
from autogoal.grammar import Boolean


class _Regex(abc.ABC):
    def __init__(self, full: Boolean):
        self.full = full
        self._name = self.__class__.__name__[:-len("Regex")].lower()

    @abc.abstractmethod
    def _regex(self):
        pass

    def run(self, input: Word()) -> Flags():
        r_exp = self._regex()
        b = re.fullmatch(r_exp, input) if self.full else re.search(r_exp, input) 
        return {f"is_{self._name}_regex": bool(b)}


@nice_repr
class UrlRegex(_Regex):
    """
    Finds if a URL is contained inside a word using regular expressions.

    ##### Examples

    ```python
    >>> regex = UrlRegex(full=True)
    >>> regex.run("https://autogoal.gitlab.io/autogoal/contributing/#license")
    {'is_url_regex': True}

    >>> regex = UrlRegex(full=True)
    >>> regex.run("There is a URL at https://autogoal.gitlab.io/autogoal/contributing/#license, who would know?")
    {'is_url_regex': False}

    >>> regex = UrlRegex(full=False)
    >>> regex.run("There is a URL at https://autogoal.gitlab.io/autogoal/contributing/#license, who would know?")
    {'is_url_regex': True}

    ```
    """
    def _regex(self):
        return r"(http|ftp|https)://([\w_-]+(?:(?:\.[\w_-]+)+))([\w.,@?^=%&:/~+#-]*[\w@?^=%&/~+#-])?"


@nice_repr
class IPRegex(_Regex):
    """
    Finds if an IP-address is contained inside a word using regular expressions.

    ##### Examples

    ```python
    >>> regex = IPRegex(full=True)
    >>> regex.run("192.168.18.1")
    {'is_ip_regex': True}

    >>> regex = IPRegex(full=True)
    >>> regex.run("There is an IP at 192.168.18.1, who would know?")
    {'is_ip_regex': False}

    >>> regex = IPRegex(full=False)
    >>> regex.run("There is an IP at 192.168.18.1, who would know?")
    {'is_ip_regex': True}

    ```
    """
    def _regex(self):
        return r"\b((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)(\.)?){4}\b"


@nice_repr
class MACRegex(_Regex):
    """
    Finds if a MAC-address is contained inside a word using regular expressions.

    ##### Examples

    ```python
    >>> regex = MACRegex(full=True)
    >>> regex.run("3D:F2:C9:A6:B3:4F")
    {'is_mac_regex': True}

    >>> regex = MACRegex(full=True)
    >>> regex.run("There is an IP at 3D-F2-C9-A6-B3-4F, who would know?")
    {'is_mac_regex': False}

    >>> regex = MACRegex(full=False)
    >>> regex.run("There is an IP at 3D:F2:C9:A6:B3:4F, who would know?")
    {'is_mac_regex': True}

    ```
    """
    def _regex(self):
        return r"([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})"


@nice_repr
class EmailRegex(_Regex):
    """
    Finds if an email is contained inside a word using regular expressions.

    ##### Examples

    ```python
    >>> regex = EmailRegex(full=True)
    >>> regex.run("someone@example.com")
    {'is_email_regex': True}

    >>> regex = EmailRegex(full=True)
    >>> regex.run("There is an email at someone@example.com, who would know?")
    {'is_email_regex': False}

    >>> regex = EmailRegex(full=False)
    >>> regex.run("There is an email at someone@example.com, who would know?")
    {'is_email_regex': True}

    ```
    """
    def _regex(self):
        return r"([a-zA-Z0-9_\-\.]+)@([a-zA-Z0-9_\-\.]+)\.([a-zA-Z]{2,5})"


@nice_repr
class PhoneRegex(_Regex):
    """
    Finds if a phone number is contained inside a word using regular expressions.

    ##### Examples

    ```python
    >>> regex = phoneRegex(full=True)
    >>> regex.run("+619123456789")
    {'is_phone_regex': True}

    >>> regex = phoneRegex(full=True)
    >>> regex.run("There is an phone at +619123456789, who would know?")
    {'is_phone_regex': False}

    >>> regex = phoneRegex(full=False)
    >>> regex.run("There is an phone at +619123456789, who would know?")
    {'is_phone_regex': True}

    ```
    """
    def _regex(self):
        return r"^((\+){1}91){1}[1-9]{1}[0-9]{9}$"



import nltk
from gensim.models.doc2vec import Doc2Vec as _Doc2Vec

from numpy import inf, nan

from autogoal.contrib.nltk._builder import NltkTokenizer, NltkTagger
from autogoal.contrib.sklearn._builder import SklearnTransformer, SklearnWrapper
from autogoal.grammar import Boolean, Categorical, Continuous, Discrete
from autogoal.kb import *
from autogoal.kb._data import *
from autogoal.utils import nice_repr

nltk.download("wordnet")
nltk.download("sentiwordnet")
nltk.download("stopwords")

from nltk.corpus import sentiwordnet as swn
from nltk.corpus import stopwords, wordnet


@nice_repr
class Doc2Vec(_Doc2Vec, SklearnTransformer):
    def __init__(
        self,
        dm: Discrete(min=0, max=2),
        dbow_words: Discrete(min=-100, max=100),
        dm_concat: Discrete(min=-100, max=100),
        dm_tag_count: Discrete(min=0, max=2),
        alpha: Continuous(min=0.001, max=0.075),
        epochs: Discrete(min=2, max=10),
        window: Discrete(min=2, max=10),
        inner_tokenizer: algorithm(Sentence(), List(Word())),
        inner_stemmer: algorithm(Word(), Stem()),
        inner_stopwords: algorithm(List(Word()), List(Word())),
        lowercase: Boolean(),
        stopwords_remove:Boolean(),
    ):

        self.inner_tokenizer = inner_tokenizer
        self.inner_stemmer = inner_stemmer
        self.inner_stopwords = inner_stopwords
        self.lowercase = lowercase
        self.stopwords_remove = stopwords_remove

        super().__init__(
            dm=dm,
            dbow_words=dbow_words,
            dm_concat=dm_concat,
            dm_tag_count=dm_tag_count,
            alpha=alpha,
            epochs=epochs,
            window=window,
        )

    def tokenize(self, sentence):
        sentence = sentence.lower() if self.lowercase else sentence
        tokens = self.inner_tokenizer.run(sentence)
        tokens = self.inner_stopwords.run(sentence) if self.stopwords_remove else tokens
        return [self.inner_stemmer.run(token) for token in tokens]

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X)

    def fit(self, X, y):
        # Data must be turned to tagged data as TaggedDocument(List(Token), Tag)
        # Tag use to be an unique integer

        from gensim.models.doc2vec import TaggedDocument as _TaggedDocument

        tagged_data = [_TaggedDocument(self.tokenize(X[i]), str(i)) for i in range(len(X))]

        self.build_vocab(tagged_data)
        return self.train(
            tagged_data, total_examples=self.corpus_count, epochs=self.epochs
        )

    def transform(self, X, y=None):
        return [self.infer_vector(x) for x in X]

    def run(self, input: List(Sentence())) -> MatrixContinuousDense():
        """This methods receive a document list and transform this into a dense continuous matrix.
       """
        return SklearnTransformer.run(self, input)


@nice_repr
class StopwordRemover:
    def __init__(
        self,
        language: Categorical(
            "danish",
            "dutch",
            "english",
            "finnish",
            "french",
            "german",
            "hungarian",
            "italian",
            "norwegian",
            "portuguese",
            "russian",
            "spanish",
            "swedish",
            "turkish",
        ),
    ):
        self.language = language
        self.words = stopwords.words(language)
        SklearnWrapper.__init__(self)

    def _train(self, input):
        return [word for word in input if word not in self.words]

    def _eval(self, input):
        return [word for word in input if word not in self.words]

    def run(self, input: List(Word())) -> List(Word()):
        """This methods receive a word list list and transform this into a word list list without stopwords.
       """
        return SklearnTransformer.run(self, input)

    def __str__(self):
        name = StopwordRemover.__name__
        return f"{name}({self.language})"


@nice_repr
class TextLowerer:
    def __init__(self):
        pass

    def fit(self, X, y=None):
        pass

    def fit_transform(self, X, y=None):
        self.fit(X, y=None)
        return self.transform(X)

    def transform(self, X, y=None):
        # Considering data as list of raw documents
        return [str.lower(x) for x in X]


@nice_repr
class WordnetConcept:
    """Find a word in Wordnet and return a List of Synset de Wordnet
    """

    def __init__(self):
        pass

    def run(self, input: Word(domain="general", language="english")) -> Synset():
        """Find a word in Wordnet and return a List of Synset de Wordnet
        """
        synsets = wordnet.synsets(input)
        names_synsets = []
        for i in synsets:
            names_synsets.append(i.name())

        return names_synsets


# @nice_repr
# class ConvertSynset2Word:
#     """Recive a Synset of nltk and return de Lemma of this
#     """

#     def __init__(self):
#         pass

#     def run(self, input: Synset(domain="general", language="english")) -> Word():
#         """Recive a Synset of nltk and return de Lemma of this
#         """
#         return Lemma(input)


@nice_repr
class SentimentWord:
    """Find a word in SentiWordnet and return a List of sentiment of the word.
    """

    def __init__(self):
        pass

    def run(self, input: Synset(domain="general", language="english")) -> Sentiment():
        """Find a word in SentiWordnet and return a List of sentiment of the word.
        """
        swn_synset = swn.senti_synset(input)

        sentiment = {}
        sentiments["positive"] = i.pos_score()
        sentiments["negative"] = i.neg_score()

        return sentiment


from nltk.chunk.named_entity import NEChunkParserTagger as _NEChunkParserTagger

@nice_repr
class NEChunkParserTagger(NltkTagger):
    def __init__(self,):
        self.tagger = _NEChunkParserTagger
        self.values = dict()

        NltkTagger.__init__(self)

    def run(self, input: List(List(Postag()))) -> List(List(Chunktag())):
        return NltkTagger.run(self, input)


@nice_repr
class GlobalChunker(SklearnWrapper):
    def __init__(
        self,
        inner_trained_pos_tagger: algorithm(List(Word()), List(Postag())),
        inner_chunker: algorithm(List(List(Postag())), List(List(Chunktag())))
    ):
        self.inner_trained_pos_tagger = inner_trained_pos_tagger
        self.inner_chunker = inner_chunker

        SklearnWrapper.__init__(self)

    def _train(self, input):
        X, y = input

        postagged_sentences = []
        for i in range(len(X)):
            sentence = []
            for itoken in range(len(X[i])):
                x_sent = X[i]
                word = x_sent[itoken]
                sentence.append(word)

            postag_sentence = self.inner_trained_pos_tagger.run((sentence, sentence))[0]
            tagged_sentence = [ (sentence[k], postag_sentence[k][1]) for k in range(len(sentence))]
            if tagged_sentence:
                postagged_sentences.append(tagged_sentence)

        tagged_sentences = []
        for i in range(len(y)):
            sentence = []
            tags = []
            for itoken in range(len(y[i])):
                y_sent = y[i]
                word, tag = y_sent[itoken]
                sentence.append(word)
                tags.append(tag)

            postag_sentence = self.inner_trained_pos_tagger.run((sentence, sentence))[0]
            tagged_sentence = [ ((sentence[k], postag_sentence[k][1]), tags[k]) for k in range(len(sentence))]
            if tagged_sentence:
                tagged_sentences.append(tagged_sentence)

        return self.inner_chunker.run((postagged_sentences, tagged_sentences))

    def _eval(self, input):
        X, y = input

        postagged_document = []
        for i in range(len(X)):
            sentence = []
            for itoken in range(len(X[i])):
                x_sent = X[i]
                word = x_sent[itoken]
                sentence.append(word)

            postag_sentence = self.inner_trained_pos_tagger.run((sentence, sentence))[0]
            tagged_sentence = [ (sentence[k], postag_sentence[k][1]) for k in range(len(sentence))]
            if tagged_sentence:
                postagged_document.append(tagged_sentence)

        return self.inner_chunker.run((postagged_document, y))

    def run(self, input: List(List(Word()))) -> List(List(Chunktag())):
        return SklearnWrapper.run(self, input)



import numpy as np
import os
from sklearn.feature_extraction import DictVectorizer
from autogoal.datasets import datapath, download


def load(representation='numeric'):
    """
    Loads corpora from [ABALONE uci dataset](https://archive.ics.uci.edu/ml/datasets/Abalone).

    ##### Examples

    ```python
    >>> X, y = load()
    >>> X.shape
    (4177, 6047)
    >>> len(y)
    4177

    ```
    """

    try:
        download("abalone")
    except:
        print("Error loading data. This may be caused due to bad connection. Please delete badly downloaded data and retry")
        raise

    path = str(datapath(os.path.dirname(os.path.abspath(__file__)))) + "/data/abalone"
    f = open(os.path.join(path, "abalone.data"), "r")

    X = []
    y = []

    for i in f.readlines():
        clean_line = i.strip().split(",")

        temp = {}
        temp["Sex"] = clean_line[0]
        temp["Length"] = clean_line[1]
        temp["Diameter"] = clean_line[2]
        temp["Height"] = clean_line[3]
        temp["Shucked weight"] = clean_line[4]
        temp["Whole weight"] = clean_line[5]
        temp["Viscera weight"] = clean_line[6]
        temp["Shell weight"] = clean_line[7]

        X.append(temp)
        y.append(clean_line[8])

    if representation == 'numeric':
        return _load_numeric(X, y)
    elif representation == 'onehot':
        return _load_onehot(X, y)

    raise ValueError("Invalid value for represenation: %s" % representation)


def _load_numeric(X, y):
    new_X = []

    for d in X:
        new_d = d.copy()

        v = d['Sex']
        if v == "M":
            new_d["Sex"] = 3
        elif v == "F":
            new_d["Sex"] = 2
        elif v == "I":
            new_d["Sex"] = 1

        new_X.append(new_d)

    return _load_onehot(new_X, y)


def _load_onehot(X, y):
    vec = DictVectorizer(sparse=False)

    return vec.fit_transform(X), np.asarray(y)



import wikipedia

from autogoal.kb import Word, Entity, Summary, Flags
from autogoal.utils import nice_repr


@nice_repr
class WikipediaSummary:
    """This class find a word in Wikipedia and return a summary in english.
    """

    def __init__(self):
        pass

    def run(self, input: Word(domain='general', language='english'))-> Summary():
        """This method use Word2Vect of gensim for tranform a word in embedding vector.
        """
        try:
            return wikipedia.summary(input)
        except:
            return ""


@nice_repr
class WikipediaContainsWord:
    """This class find a word in Wikipedia and return a summary in english.
    """

    def __init__(self):
        pass

    def run(self, input: Word(domain='general', language='english'))-> Flags():
        """This method use Word2Vect of gensim for tranform a word in embedding vector.
        """
        return dict(in_wikipedia=bool(wikipedia.search(input)))


@nice_repr
class WikipediaSummarySpanish:
    """This class find a word in Wikipedia and return a summary in Spanish.
    """

    def __init__(self):
        wikipedia.set_lang("es")

    def run(self, input: Word(domain='general', language='spanish'))-> Summary():
        """This method use Word2Vect of gensim for tranform a word in embedding vector.
        """
        try:
            return wikipedia.summary(input)
        except:
            return ""


@nice_repr
class WikipediaContainsWordSpanish:
    """This class find a word in Wikipedia and return a summary in Spanish.
    """

    def __init__(self):
        wikipedia.set_lang("es")

    def run(self, input: Word(domain='general', language='spanish'))-> Flags():
        """This method use Word2Vect of gensim for tranform a word in embedding vector.
        """
        return dict(in_wikipedia=bool(wikipedia.search(input)))



import inspect
import textwrap
import functools

from typing import Callable

from autogoal.search import PESearch
from autogoal.grammar import generate_cfg


class _ParamsDict(dict):
    pass


def _make_params_func(fn: Callable):
    signature = inspect.signature(fn)

    func_name = f"{fn.__name__}_params"
    args_names = signature.parameters.keys()

    def annotation_repr(ann):
        if inspect.isclass(ann) or inspect.isfunction(ann):
            return ann.__name__

        return repr(ann)

    args_line = ",\n                ".join(f"{k}={k}" for k in args_names)
    params_line = ", ".join(f"{arg.name}:{annotation_repr(arg.annotation)}" for arg in signature.parameters.values())

    func_code = textwrap.dedent(
        f"""
        def {func_name}({params_line}):
            return _ParamsDict(
                {args_line}
            )
        """
    )

    globals_dict = dict(fn.__globals__)
    globals_dict['_ParamsDict'] = _ParamsDict
    locals_dict = {}
    exec(func_code, globals_dict, locals_dict)
    return locals_dict[func_name]


def optimize(fn, search_strategy=PESearch, generations=100, pop_size=10, allow_duplicates=False, logger=None, **kwargs):
    """
    A general-purpose optimization function.

    Simply define any function `fn` with suitable parameter annotations
    and apply `optimize`.

    ##### Parameters

    * `search_strategy`: customize the search strategy. By default a `PESearch` will be performed.
    * `generations`: max number of generations to run.
    * `logger`: instance of `Logger` (or list) to pass to the search strategy.
    * `**kwargs`: additional keyword arguments passed to the search strategy constructor.
    """
    params_func = _make_params_func(fn)

    @functools.wraps(fn)
    def eval_func(kwargs):
        return fn(**kwargs)

    grammar = generate_cfg(params_func)

    search = search_strategy(grammar, eval_func, pop_size=pop_size, allow_duplicates=allow_duplicates, **kwargs)
    best, best_fn = search.run(generations, logger=logger)

    return best, best_fn



import gensim
import nltk
import black

import textwrap
import datetime
import inspect
import re
import numpy as np
import warnings
import abc
import importlib
import enlighten


from pathlib import Path
from autogoal.kb import *
from autogoal.grammar import Discrete, Continuous, Categorical, Boolean
from autogoal.contrib.sklearn._builder import SklearnWrapper
from ._utils import (
    _is_algorithm,
    get_input_output,
    is_algorithm,
    _is_tagger,
    is_pretrained_tagger,
)

languages = [
    "arabic",
    "danish",
    "dutch",
    "english",
    "finnish",
    "french",
    "german",
    "hungarian",
    "italian",
    "norwegian",
    "portuguese",
    "romanian",
    "russian",
    "spanish",
    "swedish",
]

languages_re = re.compile("|".join(languages))


GENERATION_RULES = dict(
    SnowballStemmer=dict(assume=True, assume_input=Word(), assume_output=Stem()),
)


class NltkTokenizer:
    def run(self, input):
        return self.tokenize(input)

    @abc.abstractmethod
    def tokenize(self, X):
        pass


class NltkStemmer:
    def run(self, input):
        return self.stem(input)

    @abc.abstractmethod
    def stem(self, X):
        pass


class NltkLemmatizer:
    def run(self, input):
        return self.lemmatize(input)

    @abc.abstractmethod
    def lemmatize(self, X):
        pass


class NltkClusterer(SklearnWrapper):
    def _train(self, input):
        X, y = input
        self.cluster(X)
        return [self.classify(x) for x in X]

    def _eval(self, input):
        X, y = input
        return X, [self.classify(x) for x in X]


class NltkTagger(SklearnWrapper):
    def _train(self, input):
        X, y = input
        tagged_sentences = [list(zip(words, tags)) for words, tags in zip(X, y)]
        self._instance = self.tagger(train=tagged_sentences)
        return X, y

    def _eval(self, input):
        X, y = input
        return [
            [tag for _, tag in sentence] for sentence in self._instance.tag_sents(X)
        ]


class NltkTrainedTagger(SklearnWrapper):
    def _train(self, input):
        X, y = input
        # X expected to be a tokenized sentence
        return self.tag(X), y

    def _eval(self, input):
        X, y = input
        # X expected to be a tokenized sentence
        return self.tag(X), y


base_classes = {
    "sent_tokenizer": "NltkTokenizer",
    "word_tokenizer": "NltkTokenizer",
    "lemmatizer": "NltkLemmatizer",
    "stemmer": "NltkStemmer",
    "word_embbeder": "SklearnWrapper",
    "doc_embbeder": "SklearnWrapper",
    "tagger": "NltkTagger",
    "trained_tagger": "NltkTrainedTagger",
}


GENERATION_RULES = dict(
    SnowballStemmer=dict(assume=True, assume_input=Word(), assume_output=Stem()),
)


def build_nltk_wrappers():
    imports = _walk(nltk)
    imports += _walk(nltk.cluster)
    imports += _walk(gensim.models)
    # imports += _walk(nltk.chunk.named_entity)
    imports += _walk(nltk.tag)

    manager = enlighten.get_manager()
    counter = manager.counter(total=len(imports), unit="classes")
    path = Path(__file__).parent / "_generated.py"

    imports = set(imports)

    with open(path, "w") as fp:
        fp.write(
            textwrap.dedent(
                f"""
            # AUTOGENERATED ON {datetime.datetime.now()}
            ## DO NOT MODIFY THIS FILE MANUALLY

            from autogoal.grammar import Continuous, Discrete, Categorical, Boolean
            from autogoal.contrib.nltk._builder import NltkStemmer, NltkTokenizer, NltkLemmatizer, NltkTagger, NltkTrainedTagger
            from autogoal.kb._data import *
            from autogoal.utils import nice_repr
            from numpy import inf, nan
            """
            )
        )

        for cls in imports:
            counter.update()
            _write_class(cls, fp)

    black.reformat_one(
        path, True, black.WriteBack.YES, black.FileMode(), black.Report()
    )

    counter.close()
    manager.stop()


def _write_class(cls, fp):
    try:
        args = _get_args(cls)
    except Exception as e:
        warnings.warn("Error to generate wrapper for %s : %s" % (cls.__name__, e))
        return

    rules = GENERATION_RULES.get(cls.__name__)
    assumed = False
    if rules:
        if rules.get("assume"):
            assumed = True
            inputs = rules.get("assume_input")
            outputs = rules.get("assume_output")

    if not assumed:
        inputs, outputs = get_input_output(cls)

    if not inputs:
        warnings.warn("Cannot find correct types for %r" % cls)
        return

    s = " " * 4
    args_str = f",\n{s * 4}".join(f"{key}: {value}" for key, value in args.items())
    init_str = f",\n{s * 5}".join(f"{key}={key}" for key in args)
    self_str = f"\n{s * 5}".join(f"self.{key}={key}" for key in args)
    values_str = f",".join(f"{key}={key}" for key in args)
    input_str, output_str = repr(inputs), repr(outputs)
    base_class = base_classes.get(is_algorithm(cls), None)  # set correct base class

    if not base_class:
        warnings.warn("Error to generate wrapper for %s" % cls.__name__)
        return

    print(cls)

    if _is_tagger(cls) and not is_pretrained_tagger(cls)[0]:
        fp.write(
            textwrap.dedent(
                f"""
            from {cls.__module__} import {cls.__name__} as _{cls.__name__}

            @nice_repr
            class {cls.__name__}({base_class}):
                def __init__(
                    self,
                    {args_str}
                ):
                    {self_str}
                    self.tagger = _{cls.__name__}
                    self.values = dict({values_str})

                    {base_class}.__init__(self)

                def run(self, input: Tuple({input_str}, {output_str})) -> {output_str}:
                    return {base_class}.run(self, input)
            """
            )
        )
    else:
        fp.write(
            textwrap.dedent(
                f"""
            from {cls.__module__} import {cls.__name__} as _{cls.__name__}

            @nice_repr
            class {cls.__name__}(_{cls.__name__}, {base_class}):
                def __init__(
                    self,
                    {args_str}
                ):
                    {self_str}
                    {base_class}.__init__(self)
                    _{cls.__name__}.__init__(
                        self,
                        {init_str}
                    )

                def run(self, input: {input_str}) -> {output_str}:
                    return {base_class}.run(self, input)
            """
            )
        )

    fp.flush()


def _write_tagger(cls, fp, args_str, init_str, input_str, output_str):
    fp.write(
        textwrap.dedent(
            f"""
        from {cls.__module__} import {cls.__name__} as _{cls.__name__}

        @nice_repr
        class {cls.__name__}(_{cls.__name__}, {base_class}):
            def __init__(
                self,
                {args_str}
            ):
                self.tagger = _{cls.__name__}
                {base_class}.__init__(self)
                _{cls.__name__}.__init__(
                    self
                    {init_str}
                )

            def run(self, input: {input_str}) -> {output_str}:
               return {base_class}.run(self, input)
        """
        )
    )

    fp.flush()


def _walk(module, name="nltk"):
    imports = []

    def _walk_p(module, name="nltk"):
        all_elements = dir(module)
        for elem in all_elements:

            if elem == "exceptions":
                continue

            name = name + "." + elem

            try:
                obj = getattr(module, elem)

                if isinstance(obj, type):
                    # ignore nltk interfaces
                    if name.endswith("I"):
                        continue

                    if not _is_algorithm(obj):
                        continue

                    imports.append(obj)

                # _walk_p(obj, name) If not module do not walk in it
            except Exception as e:
                pass

            try:
                inner_module = importlib.import_module(name)
                _walk_p(inner_module, name)
            except:
                pass

    _walk_p(module, name)

    imports.sort(key=lambda c: (c.__module__, c.__name__))
    return imports


def _find_parameter_values(parameter, cls):
    documentation = []
    lines = cls.__doc__.split("\n")

    while lines:
        l = lines.pop(0)
        if l.strip().startswith(parameter):
            documentation.append(l)
            tabs = l.index(parameter)
            break

    while lines:
        l = lines.pop(0)

        if not l.strip():
            continue

        if l.startswith(" " * (tabs + 1)):
            documentation.append(l)
        else:
            break

    options = set(re.findall(r"'(\w+)'", " ".join(documentation)))
    valid = []
    invalid = []
    skip = set(["deprecated", "auto_deprecated", "precomputed"])

    for opt in options:
        opt = opt.lower()
        if opt in skip:
            continue
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                cls(**{parameter: opt}).fit(np.ones((10, 10)), [True] * 5 + [False] * 5)
                valid.append(opt)
        except Exception as e:
            invalid.append(opt)

    return sorted(valid)


def _find_language_values(cls):
    global languages_re
    documentation = cls.__doc__
    return set(languages_re.findall(str.lower(documentation)))


def _get_args(cls):
    full_specs = inspect.getfullargspec(cls.__init__)

    args = full_specs.args
    specs = full_specs.defaults

    if not args or not specs:
        return {}

    non_kwargs = [arg for arg in args[: -len(specs) :] if arg != "self"]

    args = args[-len(specs) :]

    args_map = {k: v for k, v in zip(args, specs)}

    drop_args = [
        "url",
        "n_jobs",
        "max_iter",
        "class_weight",
        "warm_start",
        "copy_X",
        "copy_x",
        "copy",
        "eps",
        "ignore_stopwords",
        "verbose",
        "load",
    ]

    for arg in drop_args:
        args_map.pop(arg, None)

    result = {}

    for arg, value in args_map.items():
        values = _get_arg_values(arg, value, cls)
        if not values:
            continue
        result[arg] = values

    for arg in non_kwargs:
        # special handling of language
        if str.lower(arg) == "language":
            values = _find_language_values(cls)
            if values:
                result[arg] = Categorical(*values)
                continue

        if str.lower(arg) == "train" and _is_tagger(cls):
            continue

        raise Exception("No values found for positional argument %s " % (arg))
    return result


def _get_arg_values(arg, value, cls):
    if isinstance(value, bool):
        return Boolean()
    if isinstance(value, int):
        return Discrete(*_get_integer_values(arg, value, cls))
    if isinstance(value, float):
        return Continuous(*_get_float_values(arg, value, cls))
    if isinstance(value, str):
        values = _find_parameter_values(arg, cls)
        return Categorical(*values) if values else None
    return None


def _get_integer_values(arg, value, cls):
    if value == 0:
        min_val = -100
        max_val = 100
    else:
        min_val = value // 2
        max_val = 2 * value

    return min_val, max_val


def _get_float_values(arg, value, cls):
    if value == 0:
        min_val = -1
        max_val = 1
    elif 0 < value <= 0.1:
        min_val = value / 100
        max_val = 1
    elif 0 < value <= 1:
        min_val = 1e-6
        max_val = 1
    else:
        min_val = value / 2
        max_val = 2 * value

    return min_val, max_val


if __name__ == "__main__":
    build_nltk_wrappers()



try:
    import spacy
except:
    print("(!) Code in `autogoal.contrib.spacy` requires `spacy =^2.2.3`.")
    print("(!) You can install it with `pip install autogoal[spacy]`.")
    raise ImportError()


from autogoal.contrib.spacy._base import SpacyNLP



from pathlib import Path

import requests
import shutil
import numpy as np

import gensim.downloader as api
from autogoal.kb import ContinuousVector, Word
from autogoal.utils import CacheManager, nice_repr
from autogoal.datasets import download_and_save
from gensim.models import KeyedVectors
from gensim.models.fasttext import FastText, FastTextKeyedVectors


@nice_repr
class Word2VecEmbedding:
    """This class transform a word in embedding vector using Word2Vec of `gensim` (using `glove-twitter-25`).

    ##### Notes

    On the first use the model Word2Vec of gensim will be downloaded. This may take a few minutes.

    If you are using the development container the model should be already downloaded for you.
    """
    def __init__(self):
        self._model = None

    @property
    def model(self) -> KeyedVectors:
        if self._model is None:
            self._model = api.load("glove-twitter-25")

        return self._model

    def run(
        self, input: Word(domain="general", language="english")
    ) -> ContinuousVector():
        """This method use Word2Vec of gensim for tranform a word in embedding vector.
        """
        try:
            return self.model.get_vector(input)
        except:
            return np.zeros(25)


@nice_repr
class Word2VecEmbeddingSpanish:
    """This class transform a word in embedding vector using Word2Vec of `gensim` (using `Spanish 3B Word2Vec`).

    ##### Notes

    On the first use the model Word2Vec of gensim will be downloaded. This may take a few minutes.

    If you are using the development container the model should be already downloaded for you.
    """
    def __init__(self):
        self._model = None

    def _load_model(self):
        url = "https://zenodo.org/record/1410403/files/keyed_vectors.zip?download=1"
        path = Path(__file__).parent / "spanish-w2v.zip"
        kv = Path(__file__).parent / "complete.kv"

        if download_and_save(url, path):
            shutil.unpack_archive(str(path), str(path.parent))

        return KeyedVectors.load(str(kv), mmap="r")

    @property
    def model(self) -> KeyedVectors:
        if self._model is None:
            self._model = self._load_model()

        return self._model

    def run(
        self, input: Word(domain="general", language="spanish")
    ) -> ContinuousVector():
        """This method use Word2Vec of gensim for tranform a word in embedding vector.
        """
        try:
            return self.model.get_vector(input.lower())
        except KeyError:
            return np.zeros(400)


@nice_repr
class FastTextEmbeddingSpanishSUC:
    """This class transform a word in embedding vector using FastText of `gensim`.

    ##### Notes

    On the first use the model will be downloaded. This may take a few minutes.
    If you are using the development container the model should be already downloaded for you.

    ##### Examples

    >>> embedder = FastTextEmbeddingSpanishSUC()
    >>> embedder.run("algoritmo")

    """
    def __init__(self):
        self._model = None

    def _load_model(self):
        url = "https://zenodo.org/record/3234051/files/embeddings-l-model.bin?download=1"
        path = Path(__file__).parent / "fasttext-spanish-suc.bin"

        download_and_save(url, path)
        return FastText.load_fasttext_format(str(path)).wv

    @property
    def model(self) -> FastTextKeyedVectors:
        if self._model is None:
            self._model = self._load_model()

        return self._model

    def run(
        self, input: Word(domain="general", language="spanish")
    ) -> ContinuousVector():
        """This method use FastText of gensim for tranform a word in embedding vector.
        """
        try:
            return self.model.get_vector(input.lower())
        except KeyError:
            return np.zeros(300)


@nice_repr
class FastTextEmbeddingSpanishSWBC:
    """This class transform a word in embedding vector using FastText of `gensim`.

    ##### Notes

    On the first use the model will be downloaded. This may take a few minutes.
    If you are using the development container the model should be already downloaded for you.

    ##### Examples

    >>> embedder = FastTextEmbeddingSpanishSWBC()
    >>> embedder.run("algoritmo")

    """
    def __init__(self):
        self._model = None

    def _load_model(self):
        url = "http://dcc.uchile.cl/~jperez/word-embeddings/fasttext-sbwc.bin"
        path = Path(__file__).parent / "fasttext-spanish-swbc.bin"

        download_and_save(url, path)
        return FastText.load_fasttext_format(str(path)).wv

    @property
    def model(self) -> FastTextKeyedVectors:
        if self._model is None:
            self._model = self._load_model()

        return self._model

    def run(
        self, input: Word(domain="general", language="spanish")
    ) -> ContinuousVector():
        """This method use FastText of gensim for tranform a word in embedding vector.
        """
        try:
            return self.model.get_vector(input.lower())
        except KeyError:
            return np.zeros(300)


@nice_repr
class GloveEmbeddingSpanishSWBC:
    """This class transform a word in embedding vector using Glove of `gensim`.

    ##### Notes

    On the first use the model will be downloaded. This may take a few minutes.
    If you are using the development container the model should be already downloaded for you.

    ##### Examples

    >>> embedder = FastTextEmbeddingSpanishSWBC()
    >>> embedder.run("algoritmo")

    """
    def __init__(self):
        self._model = None

    def _load_model(self):
        url = "http://dcc.uchile.cl/~jperez/word-embeddings/glove-sbwc.i25.bin"
        path = Path(__file__).parent / "glove-spanish-swbc.bin"

        download_and_save(url, path)
        return FastText.load_fasttext_format(str(path)).wv

    @property
    def model(self) -> FastTextKeyedVectors:
        if self._model is None:
            self._model = self._load_model()

        return self._model

    def run(
        self, input: Word(domain="general", language="spanish")
    ) -> ContinuousVector():
        """This method use FastText of gensim for tranform a word in embedding vector.
        """
        try:
            return self.model.get_vector(input.lower())
        except KeyError:
            return np.zeros(300)



from autogoal.datasets import download, datapath

from autogoal.datasets.ehealthkd20._utils import Collection


def load_training_entities():
    # download("ehealthkd20")

    training_path = datapath("ehealthkd20") / "training"

    collection = Collection().load_dir(training_path)
    



import numpy as np
import os

from autogoal.datasets import download, datapath
from sklearn.feature_extraction import DictVectorizer


def load(representation='onehot'):
    download("uci_cars")

    f = open(datapath("uci_cars") / "car.data", "r")

    X = []
    y = []

    for i in f.readlines():
        clean_line = i.strip().split(",")

        temp = {}
        temp["buying"] = clean_line[0]
        temp["maint"] = clean_line[1]
        temp["doors"] = clean_line[2]
        temp["persons"] = clean_line[3]
        temp["lug_boot"] = clean_line[4]
        temp["safety"] = clean_line[5]

        X.append(temp)
        y.append(clean_line[6])

    if representation == 'numeric':
        return _load_numeric(X, y)
    elif representation == 'onehot':
        return _load_onehot(X, y)

    raise ValueError("Invalid value for represenation: %s" % representation)


def _load_numeric(X, y):
    new_X = []

    for d in X:
        new_d = {}
        for k, v in d.items():

            if k == "buying":
                if v == "vhigh":
                    new_d["buying"] = 4
                elif v == "high":
                    new_d["buying"] = 3
                elif v == "med":
                    new_d["buying"] = 2
                elif v == "low":
                    new_d["buying"] = 1

            if k == "maint":
                if v == "vhigh":
                    new_d["maint"] = 4
                elif v == "high":
                    new_d["maint"] = 3
                elif v == "med":
                    new_d["maint"] = 2
                elif v == "low":
                    new_d["maint"] = 1

            if k == "doors":
                if v == "5more":
                    new_d["doors"] = 4
                elif v == "4":
                    new_d["doors"] = 3
                elif v == "3":
                    new_d["doors"] = 2
                elif v == "2":
                    new_d["doors"] = 1

            if k == "persons":
                if v == "more":
                    new_d["persons"] = 3
                elif v == "4":
                    new_d["persons"] = 2
                elif v == "2":
                    new_d["persons"] = 1

            if k == "lug_boot":
                if v == "big":
                    new_d["lug_boot"] = 3
                elif v == "med":
                    new_d["lug_boot"] = 2
                elif v == "small":
                    new_d["lug_boot"] = 1

            if k == "safety":
                if v == "high":
                    new_d["safety"] = 3
                elif v == "med":
                    new_d["safety"] = 2
                elif v == "low":
                    new_d["safety"] = 1

        new_X.append(new_d)

    return _load_onehot(X, y)


def _load_onehot(X, y):
    vec = DictVectorizer(sparse=False)

    return vec.fit_transform(X), np.asarray(y)



import inspect
import warnings
import networkx as nx
from collections import namedtuple, defaultdict
import logging

from autogoal.grammar import GraphSpace, Graph, generate_cfg, CfgInitializer
from autogoal.utils import nice_repr
from autogoal.kb._data import (
    conforms,
    build_composite_tuple,
    make_list_wrapper,
    Tuple,
    build_composite_list,
    List,
    DataType,
    DATA_TYPES,
)


MAX_LIST_DEPTH = 5


@nice_repr
class PipelineNode:
    def __init__(self, algorithm, input, output):
        self.algorithm = algorithm
        self.input = tuple(input)
        self.output = tuple(output)
        self.grammar = generate_cfg(algorithm)

    def __eq__(self, other: "PipelineNode"):
        if not isinstance(other, PipelineNode):
            return False
                

        return self.algorithm == other.algorithm and self.input == other.input and self.output == other.output

    def __hash__(self):
        return hash(self.algorithm) ^ hash(self.input) ^ hash(self.output)

    @property
    def __name__(self):
        return self.algorithm.__name__

    def sample(self, sampler):
        return self.grammar.sample(sampler=sampler)


@nice_repr
class PipelineStart:
    __name__ = "PipelineStart"

    def __init__(self, types):
        self.input = types
        self.output = types

    def sample(self, sampler):
        return self


@nice_repr
class PipelineEnd:
    __name__ = "PipelineEnd"

    def __init__(self, types):
        self.input = types
        self.output = types

    def sample(self, sampler):
        return self


def build_pipeline_graph(input: DataType, output: DataType, registry, max_list_depth=3, max_pipeline_width=3) -> "PipelineBuilder":
    """
    Creates a `PipelineBuilder` instance that generates all pipelines
    from `input` to `output` types.

    ##### Parameters

    - `input`: type descriptor for the desired input.
    - `output`: type descriptor for the desired output.
    - `registry`: list of available classes to build the pipelines.
    """
    
    logger = logging.getLogger("autogoal.build_pipeline_graph")
    
    # First we will unpack the input and output type and
    # store them in actual lists for easier use

    if isinstance(input, Tuple):
        input_type = list(input.inner)
    else:
        input_type = [input]

    if isinstance(output, Tuple):
        output_type = list(output.inner)
    else:
        output_type = [output]

    logger.info(f"input_type={input_type}")
    logger.info(f"output_type={output_type}")

    # Before starting, let's create all the List[T] wrappers up to 
    # `max_list_depth` and add them to `registry`, so that they are available later
    for algorithm in list(registry):
        for _ in range(max_list_depth):
            algorithm = make_list_wrapper(algorithm)
            registry.append(algorithm)

    # We will also need an index to quickly find out which algorithms
    # accept each input type
    index = defaultdict(set)

    for algorithm in registry:
        types = _get_annotations(algorithm).input
        types = list(types.inner) if isinstance(types, Tuple) else [types]

        for t in types:
            index[t].add(algorithm)

    logger.info(f"Built algorithm index with {len(index)} entries and {len(registry)} total algorithms.")

    # The graph contains all the algorithms, each algorithm is connected
    # to all those nodes that it can process, which are nodes whose output
    # type is a superset of what the algorithm requires.
    G = Graph()

    # For each node stored in the graph, we will store also the full list
    # of all inputs and outputs that we can guarantee are available at this point.
    # Initially we add the `Start` node, which produces all of the inputs,
    # and the `End` node which consumes all the outputs.
    start_node = PipelineStart(input_type)
    end_node = PipelineEnd(output_type)
    G.add_edge(GraphSpace.Start, start_node)
    G.add_edge(end_node, GraphSpace.End)

    # We will apply a BFS algorithm at this point. We will make sure
    # that once a node is processed, all the algorithms to which it could
    # potentially connect are stored in the graph.
    # Initially the `Start` node is the only one open.
    open_nodes = [start_node]
    closed_nodes = set()

    while open_nodes:
        # This is the next node we will need to connect.
        node = open_nodes.pop(0)

        if node in closed_nodes:
            continue

        # When leaving this node we can guarantee that we have the types in this list.
        types = node.output
        logger.info(f"Processing node={node}")

        # We will need this method to check if all of the input types of and algorithm are
        # guaranteed at this point, i.e., if they are available in `types`,
        # or at least a conforming type is.
        def type_is_guaranteed(input_type):
            for other_type in types:
                if conforms(other_type, input_type):
                    return True

            return False

        # In this point we have to identify all the algorithms that could continue
        # from this point on. These are all the algorithms whose input expects a subset
        # of the types that we already have.
        potential_algorithms = set()

        for t in types:
            potential_algorithms |= index[t]

        for algorithm in potential_algorithms:
            annotations = _get_annotations(algorithm)
            algorithm_input_types = list(annotations.input.inner) if isinstance(annotations.input, Tuple) else [annotations.input]
            algorithm_output_types = list(annotations.output.inner) if isinstance(annotations.output, Tuple) else [annotations.output]
            logger.info(f"Analyzing algorithm={algorithm.__name__} with inputs={algorithm_input_types} and outputs={algorithm_output_types}")

            if any(not type_is_guaranteed(input_type) for input_type in algorithm_input_types):
                logger.info(f"Skipping algorithm={algorithm.__name__}")
                continue
                    
            # At this point we can add the current algorithm to the graph.
            # First, we make the current algorithm "consume" the input types,
            # hence, the output types produced at this point are the output types
            # this algorithm provides plus any input type not consumed so far.
            output_types = sorted(set([t for t in types if t not in algorithm_input_types] + algorithm_output_types), key=str)

            if len(output_types) > max_pipeline_width:
                continue
            
            # We add this node to the graph and we mark that it consumes the inputs,
            # so that later when sampling we can correctly align all the types.
            # When building the node, we can get a `ValueError` if the internal
            # grammar cannot be built; in that case, we simply skip it
            try:
                new_node = PipelineNode(algorithm=algorithm, input=types, output=output_types)
                G.add_node(new_node)
                G.add_edge(node, new_node)
                open_nodes.append(new_node)
                logger.info(f"Adding node={algorithm.__name__} producing types={output_types}")
            except ValueError as e:
                logger.warning(f"Node={algorithm.__name__} cannot be built. Error={e}.")           

        # Let's check if we can add the `End` node.
        if all(type_is_guaranteed(t) for t in output_type):
            G.add_edge(node, end_node)
            logger.info("Connecting to end node")
            
        closed_nodes.add(node)

    # Once done we have to check if the `End` node was at some point included in the graph.
    # Otherwise that means there is no possible path.
    if GraphSpace.End not in G:
        raise TypeError(
            "No pipelines can be constructed from input:%r to output:%r."
            % (input, output)
        )

    # Now we remove all nodes that don't participate in any path
    # leaving to `End`
    reachable_from_end = set(nx.dfs_preorder_nodes(G.reverse(False), GraphSpace.End))
    unreachable_nodes = set(G.nodes) - reachable_from_end
    G.remove_nodes_from(unreachable_nodes)

    # If the node `Start` was removed, that means the graph is disconnected.
    if not GraphSpace.Start in G:
        raise TypeError(
            "No pipelines can be constructed from input:%r to output:%r."
            % (input, output)
        )

    return PipelineBuilder(G, registry)
                    

def build_pipelines(input, output, registry) -> "PipelineBuilder":
    """
    Creates a `PipelineBuilder` instance that generates all pipelines
    from `input` to `output` types.

    ##### Parameters

    - `input`: type descriptor for the desired input.
    - `output`: type descriptor for the desired output.
    - `registry`: list of available classes to build the pipelines.
    """

    # warnings.warn(
    #     "This method is deprecated and not under use by AutoGOAL's"
    #     " internal API anymore, use `build_pipeline_graph` instead.",
    #     category=DeprecationWarning,
    #     stacklevel=2,
    # )

    list_pairs = set()
    types_queue = []

    if isinstance(input, Tuple):
        types_queue.extend(input.inner)
    else:
        types_queue.append(input)

    if isinstance(output, Tuple):
        types_queue.extend(output.inner)
    else:
        types_queue.append(output)

    types_seen = set()

    while types_queue:
        output_type = types_queue.pop(0)

        def build(internal_output, depth):
            if internal_output in types_seen:
                return

            for other_clss in registry:
                annotations = _get_annotations(other_clss)

                if annotations in list_pairs:
                    continue

                other_input = annotations.input
                other_output = annotations.output

                if other_input == other_output:
                    continue

                if not conforms(internal_output, other_input):
                    continue

                other_wrapper = build_composite_list(other_input, other_output, depth)
                list_pairs.add(annotations)
                registry.append(other_wrapper)
                types_queue.append(_get_annotations(other_wrapper).output)

        depth = 0

        while isinstance(output_type, List):
            if output_type.depth() >= MAX_LIST_DEPTH:
                break

            depth += 1

            output_type = output_type.inner
            build(output_type, depth)
            types_seen.add(output_type)

            print(output_type)

    list_tuples = set()

    def connect_tuple_wrappers(node, output_type):
        if not isinstance(output_type, Tuple):
            return

        for index in range(0, len(output_type.inner)):
            internal_input = output_type.inner[index]

            for other_clss in registry:
                annotations = _get_annotations(other_clss)
                other_input = annotations.input

                if not (conforms(internal_input, other_input) and other_clss != node):
                    continue

                # `other_class` has input compatible with one element in the Tuple
                # build the output `Tuple[..., internal_output, ...]` of the wrapper class
                internal_output = annotations.output
                output_tuple = list(output_type.inner)
                output_tuple[index] = internal_output
                output_tuple_type = Tuple(*output_tuple)

                # dynamic class representing the wrapper algorithm
                if (index, output_type, output_tuple_type) in list_tuples:
                    continue

                other_wrapper = build_composite_tuple(
                    index, output_type, output_tuple_type
                )
                list_tuples.add((index, output_type, output_tuple_type))
                registry.append(other_wrapper)

                open_nodes.append(other_wrapper)
                G.add_edge(node, other_wrapper)

    G = Graph()

    open_nodes = []
    closed_nodes = set()

    # Enqueue open nodes
    for clss in registry:
        if conforms(input, _get_annotations(clss).input):
            open_nodes.append(clss)
            G.add_edge(GraphSpace.Start, clss)

    connect_tuple_wrappers(GraphSpace.Start, input)

    if GraphSpace.Start not in G:
        raise TypeError("There are no classes compatible with input type:%r." % input)

    while open_nodes:
        clss = open_nodes.pop(0)

        if clss in closed_nodes:
            continue

        closed_nodes.add(clss)
        output_type = _get_annotations(clss).output

        for other_clss in registry:
            other_input = _get_annotations(other_clss).input
            if conforms(output_type, other_input) and other_clss != clss:
                open_nodes.append(other_clss)
                G.add_edge(clss, other_clss)

        connect_tuple_wrappers(clss, output_type)

        if conforms(output_type, output):
            G.add_edge(clss, GraphSpace.End)

    if GraphSpace.End not in G:
        raise TypeError(
            "No pipelines can be constructed from input:%r to output:%r."
            % (input, output)
        )

    reachable_from_end = set(nx.dfs_preorder_nodes(G.reverse(False), GraphSpace.End))
    unreachable_nodes = set(G.nodes) - reachable_from_end
    G.remove_nodes_from(unreachable_nodes)

    if not GraphSpace.Start in G:
        raise TypeError(
            "No pipelines can be constructed from input:%r to output:%r."
            % (input, output)
        )

    return PipelineBuilder(G, registry)


@nice_repr
class Pipeline:
    def __init__(self, steps):
        self.steps = steps

    def send(self, msg: str, *args, **kwargs):
        found = False

        for step in self.steps:
            if hasattr(step, msg):
                getattr(step, msg)(*args, **kwargs)
                found = True
            elif hasattr(step, "send"):
                step.send(msg, *args, **kwargs)
                found = True

        if not found:
            warnings.warn(f"No step answered message {msg}.")

    def run(self, x):
        for step in self.steps:
            try:
                x = step.run(x)
            except Exception as e:
                raise e from None

        return x


class NewPipelineBuilder(GraphSpace):
    def __init__(self, graph, registry):
        super().__init__(graph, initializer=self.initialize)

    def initialize(self, node, sampler):
        return node.sample(sampler)

    def sample(self, *args, **kwargs) -> Pipeline:
        path = super().sample(*args, **kwargs)
        return Pipeline(path)


class PipelineBuilder(GraphSpace):
    def __init__(self, graph, registry):
        super().__init__(graph, initializer=CfgInitializer(registry=registry))

    def sample(self, *args, **kwargs) -> Pipeline:
        path = super().sample(*args, **kwargs)
        return Pipeline(path)


Annotations = namedtuple("Annotations", ["input", "output"])


def _get_annotations(clss):
    run_method = clss.run
    input_type = inspect.signature(run_method).parameters["input"].annotation
    output_type = inspect.signature(run_method).return_annotation

    return Annotations(input=input_type, output=output_type)


def _has_input(clss, input):
    return conforms(input, _get_annotations(clss).input)


def _has_output(clss, output):
    return conforms(_get_annotations(clss).output, output)



# coding: utf8

import bisect
import collections
import re
import warnings
from collections import defaultdict
from pathlib import Path
from typing import List

from autogoal.datasets.ehealthkd20._tools import (
    AnnFile,
    AttributeAnnotation,
    EntityAnnotation,
    EventAnnotation,
    RelationAnnotation,
    SameAsAnnotation,
)

ENTITIES = ["Concept", "Action", "Predicate", "Reference"]

RELATIONS = [
    "is-a",
    "same-as",
    "part-of",
    "has-property",
    "causes",
    "entails",
    "in-context",
    "in-place",
    "in-time",
    "subject",
    "target",
    "domain",
    "arg",
]


class Keyphrase:
    def __init__(self, sentence, label, id, spans):
        self.sentence: Sentence = sentence
        self.label = label
        self.id = id
        self.spans = spans
        self.attributes: List[Attribute] = []

    def split(self):
        if len(self.spans) > 1:
            raise TypeError("Cannot split a keyphrase with multiple spans")

        start, end = self.spans[0]
        spans = []
        spans.append(start)

        for i, c in enumerate(self.text):
            if c == " ":
                spans.append(start + i)
                spans.append(start + i + 1)

        spans.append(end)
        self.spans = [(spans[i], spans[i + 1]) for i in range(0, len(spans), 2)]

    def clone(self, sentence, shallow=False) -> "Keyphrase":
        k = Keyphrase(sentence, self.label, self.id, self.spans)
        k.attributes = [a if shallow else a.clone(k) for a in self.attributes]
        return k

    @property
    def text(self):
        return " ".join(self.sentence.text[s:e] for (s, e) in self.spans)

    def __repr__(self):
        return "Keyphrase(text=%r, label=%r, id=%r, attr=%r)" % (
            self.text,
            self.label,
            self.id,
            self.attributes,
        )

    def as_ann(self, shift):
        return "T{0}\t{1} {2}\t{3}\n".format(
            self.id,
            self.label,
            ";".join(
                "{} {}".format(start + shift, end + shift) for start, end in self.spans
            ),
            self.text,
        )

    def matches(self, other: "Keyphrase", label=None):
        return (
            isinstance(other, Keyphrase)
            and self.sentence.text == other.sentence.text
            and self.spans == other.spans
            and (
                (label is None and self.label == other.label)
                or (label is not None and self.label == label)
            )
        )

    def find_attributes(self, label) -> "Attribute":
        return [attr for attr in self.attributes if attr.label == label]


class Relation:
    def __init__(self, sentence, origin, destination, label):
        self.sentence = sentence
        self.origin = origin
        self.destination = destination
        self.label = label

    def clone(self, sentence) -> "Relation":
        return Relation(sentence, self.origin, self.destination, self.label)

    @property
    def from_phrase(self) -> Keyphrase:
        return self.sentence.find_keyphrase(id=self.origin)

    @property
    def to_phrase(self) -> Keyphrase:
        return self.sentence.find_keyphrase(id=self.destination)

    class _Unk:
        text = "UNK"

    def __repr__(self):
        from_phrase = (self.from_phrase or Relation._Unk()).text
        to_phrase = (self.to_phrase or Relation._Unk()).text
        return "Relation(from=%r, to=%r, label=%r)" % (
            from_phrase,
            to_phrase,
            self.label,
        )

    def as_ann(self, shift):
        if self.label == "same-as":
            return "*\tsame-as T{0} T{1}\n".format(self.origin, self.destination)
        else:
            return "R{0}\t{1} Arg1:T{2} Arg2:T{3}\n".format(
                shift, self.label, self.origin, self.destination
            )

    def matches(self, other: "Relation", label=None):
        return (
            isinstance(other, Relation)
            and self.sentence.text == other.sentence.text
            and self.from_phrase.matches(other.from_phrase)
            and self.to_phrase.matches(other.to_phrase)
            and (
                (label is None and self.label == other.label)
                or (label is not None and self.label == label)
            )
        )


class Attribute:
    def __init__(self, keyphrase: Keyphrase, label):
        self.keyphrase = keyphrase
        self.label = label

    def clone(self, keyphrase) -> "Attribute":
        return Attribute(keyphrase, self.label)

    def __repr__(self):
        return "Attribute(label=%r)" % (self.label,)

    def as_ann(self, shift):
        return "A{0}\t{1} T{2}\n".format(shift, self.label, self.keyphrase.id)


class Sentence:
    def __init__(self, text):
        self.text = text
        self.keyphrases: List[Keyphrase] = []
        self.relations: List[Relation] = []

    def clone(self, shallow=False) -> "Sentence":
        s = Sentence(self.text)
        s.keyphrases = [k if shallow else k.clone(s) for k in self.keyphrases]
        s.relations = [r if shallow else r.clone(s) for r in self.relations]
        return s

    def fix_ids(self, start=1):
        next_id = start

        copy = self.clone()
        for k, kc in zip(self.keyphrases, copy.keyphrases):
            for r, rc in zip(self.relations, copy.relations):
                if rc.origin == kc.id:
                    r.origin = next_id
                if rc.destination == kc.id:
                    r.destination = next_id

            k.id = next_id
            next_id += 1

        return next_id

    def overlapping_keyphrases(self):
        result = []

        for s1 in self.keyphrases:
            overlaps = set([s1])

            for s2 in self.keyphrases:
                if s2.spans == s1.spans:
                    overlaps.add(s2)

            if len(overlaps) > 1 and overlaps not in result:
                result.append(overlaps)

        return result

    def merge_overlapping_keyphrases(self):
        overlaps = self.overlapping_keyphrases()

        for keyphrases in overlaps:
            keyphrases = list(keyphrases)
            first = keyphrases[0]
            rest = keyphrases[1:]
            rest_ids = [k.id for k in rest]

            for relation in self.relations:
                if relation.origin in rest_ids:
                    print(
                        "Changing %r origin from %s to %s"
                        % (relation, relation.origin, first.id)
                    )
                    relation.origin = first.id
                if relation.destination in rest_ids:
                    print(
                        "Changing %r destination from %s to %s"
                        % (relation, relation.destination, first.id)
                    )
                    relation.destination = first.id

            for keyp in rest:
                self.keyphrases.remove(keyp)

    def dup_relations(self):
        dup_relations = collections.defaultdict(lambda: [])

        for r in self.relations:
            dup_relations[(r.label, r.origin, r.destination)].append(r)

        return {k: v for k, v in dup_relations.items() if len(v) > 1}

    def remove_dup_relations(self):
        new_relations = {}

        for r in self.relations:
            new_relations[(r.label, r.origin, r.destination)] = r

        self.relations = list(new_relations.values())

    def find_first_match(self, annotation, label=None):
        matches = self.find_matches(annotation, label)
        return None if not matches else matches[0]

    def find_matches(self, annotation, label=None):
        if isinstance(annotation, Keyphrase):
            return [k for k in self.keyphrases if k.matches(annotation, label)]
        elif isinstance(annotation, Relation):
            return [r for r in self.relations if r.matches(annotation, label)]
        else:
            raise TypeError("Invalid annotation")

    def find_keyphrase(self, id=None, start=None, end=None, spans=None) -> Keyphrase:
        if id is not None:
            return self._find_keyphrase_by_id(id)
        if spans is None:
            spans = [(start, end)]
        return self._find_keyphrase_by_spans(spans)

    def find_relations(self, orig, dest) -> List[Relation]:
        results = []

        for r in self.relations:
            if r.origin == orig and r.destination == dest:
                results.append(r)

        return results

    def find_relation(self, orig, dest, label) -> Relation:
        for r in self.relations:
            if r.origin == orig and r.destination == dest and label == r.label:
                return r

        return None

    def _find_keyphrase_by_id(self, id) -> Keyphrase:
        for k in self.keyphrases:
            if k.id == id:
                return k

        return None

    def _find_keyphrase_by_spans(self, spans) -> Keyphrase:
        for k in self.keyphrases:
            if k.spans == spans:
                return k

        return None

    def sort(self):
        self.keyphrases.sort(
            key=lambda k: tuple([s for s, e in k.spans] + [e for s, e in k.spans])
        )

    def __len__(self):
        return len(self.text)

    def __repr__(self):
        return "Sentence(text=%r, keyphrases=%r, relations=%r)" % (
            self.text,
            self.keyphrases,
            self.relations,
        )

    @staticmethod
    def load(finput) -> "List[Sentence]":
        return [
            Sentence(s.strip())
            for s in finput.read_text(encoding="utf8").splitlines()
            if s
        ]

    @property
    def annotated(self):
        return self.keyphrases or self.relations


class Collection:
    def __init__(self, sentences=None):
        self.sentences: List[Sentence] = sentences or []

    def clone(self, skip_empty=False) -> "Collection":
        return Collection(
            [s.clone() for s in self.sentences if not skip_empty or s.annotated]
        )

    def merge(self, *collections: "Collection", skip_empty=False):
        clone = self.clone(skip_empty)
        sentences = [
            s.clone()
            for c in collections
            for s in c.sentences
            if not skip_empty or s.annotated
        ]
        clone.sentences.extend(sentences)
        return clone

    def __len__(self):
        return len(self.sentences)

    def fix_ids(self):
        next_id = 1

        for s in self.sentences:
            next_id = s.fix_ids(next_id)

    def filter(self, keyphrase=None, relation=None, attribute=None,) -> "Collection":
        sentences = []
        for sentence in self.sentences:
            s = Sentence(sentence.text)

            # keyphrases
            s.keyphrases = [
                k.clone(s)
                for k in sentence.keyphrases
                if keyphrase is None or keyphrase(k)
            ]

            # attributes
            if attribute is not None:
                for k in s.keyphrases:
                    k.attributes = [
                        a for a in k.attributes if attribute is None or attribute(a)
                    ]
                s.keyphrases = [k for k in s.keyphrases if k.attributes]

            # relations
            s.relations = [
                r.clone(s)
                for r in sentence.relations
                if (relation is None or relation(r))
                and (
                    keyphrase is None
                    or (keyphrase(r.from_phrase) and keyphrase(r.to_phrase))
                )
                and (
                    attribute is None
                    or (
                        any(attribute(a) for a in r.from_phrase.attributes)
                        and any(attribute(a) for a in r.to_phrase.attributes)
                    )
                )
            ]
            sentences.append(s)
        return Collection(sentences)

    def filter_keyphrase(self, labels) -> "Collection":
        return self.filter(keyphrase=lambda k: k.label in labels)

    def filter_relation(self, labels) -> "Collection":
        return self.filter(relation=lambda r: r.label in labels)

    def filter_attribute(self, labels) -> "Collection":
        return self.filter(attribute=lambda a: a.label in labels)

    def find_first_match(self, text) -> Sentence:
        matches = self.find_matches(text)
        return None if not matches else matches[0]

    def find_matches(self, text) -> List[Sentence]:
        return [s for s in self.sentences if s.text == text]

    def load(
        self,
        finput: Path,
        *,
        legacy=True,
        keyphrases=True,
        relations=True,
        attributes=True
    ) -> "Collection":
        return CollectionV2Handler.load(
            self,
            finput,
            legacy=legacy,
            keyphrases=keyphrases,
            relations=relations,
            attributes=attributes,
        )

    def dump(self, text_file: Path, skip_empty_sentences=True):
        return CollectionV2Handler.dump(self, text_file, skip_empty_sentences)

    def load_dir(
        self,
        finput: Path,
        *,
        legacy=True,
        keyphrases=True,
        relations=True,
        attributes=True
    ) -> "Collection":
        return CollectionV2Handler.load_dir(
            self,
            finput,
            legacy=legacy,
            keyphrases=keyphrases,
            relations=relations,
            attributes=attributes,
        )


class CollectionHandler:
    @classmethod
    def load_dir(cls, collection: Collection, finput: Path, **kargs) -> Collection:
        pass

    @classmethod
    def load(cls, collection: Collection, finput: Path, **kargs) -> Collection:
        pass

    @classmethod
    def dump(cls, collection: Collection, text_file: Path, skip_empty_sentences=True):
        pass


class CollectionV1Handler(CollectionHandler):
    @classmethod
    def load_dir(cls, collection: Collection, finput: Path) -> Collection:
        for item in finput.iterdir():
            if re.fullmatch(r".*put_scenario.*\.txt", item.name):
                cls.load(collection, item)
        return collection

    @classmethod
    def load(cls, collection: Collection, finput: Path) -> Collection:
        input_b_file = finput.parent / ("output_b_" + finput.name.split("_")[1])

        sentence_by_id = cls._load_keyphrases(collection, finput)

        for line in input_b_file.open(encoding="utf8").readlines():
            label, src, dst = line.strip().split("\t")
            src, dst = int(src), int(dst)

            the_sentence = sentence_by_id[src]

            if the_sentence != sentence_by_id[dst]:
                warnings.warn(
                    "In file '%s' relation '%s' between %i and %i crosses sentence boundaries and has been ignored."
                    % (finput, label, src, dst)
                )
                continue

            assert sentence_by_id[dst] == the_sentence

            the_sentence.relations.append(
                Relation(the_sentence, src, dst, label.lower())
            )

        return collection

    @classmethod
    def _load_keyphrases(cls, collection: Collection, finput: Path):
        cls._load_input(collection, finput)

        input_a_file = finput.parent / ("output_a_" + finput.name.split("_")[1])

        sentences_length = [len(s.text) for s in collection.sentences]
        for i in range(1, len(sentences_length)):
            sentences_length[i] += sentences_length[i - 1] + 1

        sentence_by_id = {}

        for line in input_a_file.open(encoding="utf8").readlines():
            lid, spans, label, _ = line.strip().split("\t")
            lid = int(lid)

            spans = [s.split() for s in spans.split(";")]
            spans = [(int(start), int(end)) for start, end in spans]

            # find the sentence where this annotation is
            i = bisect.bisect(sentences_length, spans[0][0])
            # correct the annotation spans
            if i > 0:
                spans = [
                    (
                        start - sentences_length[i - 1] - 1,
                        end - sentences_length[i - 1] - 1,
                    )
                    for start, end in spans
                ]
                spans.sort(key=lambda t: t[0])
            # store the annotation in the corresponding sentence
            the_sentence = collection.sentences[i]
            keyphrase = Keyphrase(the_sentence, label, lid, spans)
            the_sentence.keyphrases.append(keyphrase)

            if len(keyphrase.spans) == 1:
                keyphrase.split()

            sentence_by_id[lid] = the_sentence

        return sentence_by_id

    @classmethod
    def _load_input(cls, collection: Collection, finput: Path):
        sentences = [s.strip() for s in finput.open(encoding="utf8").readlines() if s]
        sentences_obj = [Sentence(text) for text in sentences]
        collection.sentences.extend(sentences_obj)

    @classmethod
    def dump(cls, collection: Collection, text_file: Path, skip_empty_sentences=True):
        collection.fix_ids()

        input_file = text_file.open("w", encoding="utf8")
        output_a_file = (
            text_file.parent / ("output_a_" + text_file.name.split("_")[1])
        ).open("w", encoding="utf8")
        output_b_file = (
            text_file.parent / ("output_b_" + text_file.name.split("_")[1])
        ).open("w", encoding="utf8")

        shift = 0

        for sentence in collection.sentences:
            if (
                not sentence.keyphrases
                and not sentence.relations
                and skip_empty_sentences
            ):
                continue

            input_file.write("{}\n".format(sentence.text))

            for keyphrase in sentence.keyphrases:
                output_a_file.write(
                    "{0}\t{1}\t{2}\t{3}\n".format(
                        keyphrase.id,
                        ";".join(
                            "{} {}".format(start + shift, end + shift)
                            for start, end in keyphrase.spans
                        ),
                        keyphrase.label,
                        keyphrase.text,
                    )
                )

            for relation in sentence.relations:
                output_b_file.write(
                    "{0}\t{1}\t{2}\n".format(
                        relation.label, relation.origin, relation.destination
                    )
                )

            shift += len(sentence) + 1


class CollectionV2Handler(CollectionHandler):
    @classmethod
    def load_dir(
        cls,
        collection: Collection,
        finput: Path,
        *,
        legacy=True,
        keyphrases=True,
        relations=True,
        attributes=True
    ) -> Collection:
        for item in finput.iterdir():
            if item.suffix == ".txt":
                cls.load(
                    collection,
                    item,
                    legacy=legacy,
                    keyphrases=keyphrases,
                    relations=relations,
                    attributes=attributes,
                )

        return collection

    @classmethod
    def load(
        cls,
        collection: Collection,
        finput: Path,
        *,
        legacy=True,
        keyphrases=True,
        relations=True,
        attributes=True
    ) -> "Collection":

        # add sentences from input .txt to Collection
        sentences = cls._load_input(collection, finput)

        # if keyphrases won't be loaded finish right there
        if not keyphrases:
            return collection

        # else, parse .ann file to start the annotation of sentences
        ann_file = cls._load_ann(finput)

        def add_relation(source_id, destination_id, ann_type, id_to_keyphrase):
            source = id_to_keyphrase[source_id]
            destination = id_to_keyphrase[destination_id]
            if source.sentence != destination.sentence:
                warnings.warn(
                    "In file '%s' relation '%s' between %i and %i crosses sentence boundaries and has been ignored."
                    % (finput, ann_type, source_id, destination_id)
                )
            else:
                relation = Relation(
                    source.sentence, source.id, destination.id, ann_type
                )
                source.sentence.relations.append(relation)

        def legacy_load(ann_file, sentences, id_to_keyphrase):
            for ann in ann_file.annotations:
                if isinstance(ann, EventAnnotation):
                    id_to_keyphrase[ann.id] = id_to_keyphrase[ann.ref]

            for ann in ann_file.annotations:
                if not isinstance(ann, EventAnnotation):
                    continue
                for label, destination in ann.args.items():
                    label = "".join(i for i in label if not i.isdigit()).lower()
                    add_relation(ann.ref, destination, label, id_to_keyphrase)

        # compute sentences' boundaries
        sentences_length = [len(s) for s in sentences]
        for i in range(1, len(sentences_length)):
            sentences_length[i] += sentences_length[i - 1] + 1

        # load keyphrases from Entity Annotations
        id_to_keyphrase = {}
        for ann in ann_file.annotations:
            if isinstance(ann, EntityAnnotation):
                tid = int(ann.id[1:])
                spans = [(int(start), int(end)) for start, end in ann.spans]
                sid, spans = cls._get_relative_ann(spans, sentences_length)
                sentence = sentences[sid]
                keyphrase = Keyphrase(sentence, ann.type, tid, spans)
                sentence.keyphrases.append(keyphrase)
                if len(keyphrase.spans) == 1:
                    keyphrase.split()
                id_to_keyphrase[ann.id] = keyphrase

        # load relations from Event Annotations (legacy support)
        if legacy and relations:
            legacy_load(ann_file, sentences, id_to_keyphrase)

        # load standard relations and attributes
        for ann in ann_file.annotations:
            if isinstance(ann, RelationAnnotation) and relations:
                add_relation(ann.arg1, ann.arg2, ann.type, id_to_keyphrase)

            elif isinstance(ann, SameAsAnnotation) and relations:
                source = ann.args[0]
                for destination in ann.args[1:]:
                    add_relation(source, destination, ann.type, id_to_keyphrase)

            elif isinstance(ann, AttributeAnnotation) and attributes:
                keyphrase = id_to_keyphrase[ann.ref]
                attribute = Attribute(keyphrase, ann.type)
                keyphrase.attributes.append(attribute)

            elif not (
                isinstance(ann, EntityAnnotation)
                or legacy
                and isinstance(ann, EventAnnotation)
            ):
                warnings.warn(
                    "In file '%s' annotation '%s' has been ignored." % (finput, ann)
                )

        for s in sentences:
            s.sort()
        return collection

    @classmethod
    def _load_input(cls, collection: Collection, finput: Path) -> List[Sentence]:
        sentences = Sentence.load(finput)
        collection.sentences.extend(sentences)
        return sentences

    @classmethod
    def _load_ann(cls, finput: Path) -> AnnFile:
        ann_path: Path = finput.parent / (finput.stem + ".ann")
        return AnnFile().load(ann_path)

    @classmethod
    def _get_relative_ann(cls, spans, sentences_length: List[int]) -> int:
        # find the sentence where this annotation is
        i = bisect.bisect(sentences_length, spans[0][0])
        # correct the annotation spans
        if i > 0:
            spans = [
                (
                    start - sentences_length[i - 1] - 1,
                    end - sentences_length[i - 1] - 1,
                )
                for start, end in spans
            ]
            spans.sort(key=lambda t: t[0])
        return i, spans

    @classmethod
    def dump(cls, collection: Collection, text_file, skip_empty_sentences=True):
        ann_path: Path = text_file.parent / (text_file.stem + ".ann")
        cls._dump_input(collection, text_file, skip_empty_sentences)
        cls._dump_ann(collection, ann_path, skip_empty_sentences)

    @classmethod
    def _dump_input(
        cls, collection: Collection, text_file: Path, skip_empty_sentences=True
    ):
        text_file.parent.mkdir(parents=True, exist_ok=True)
        text_file.write_text(
            "\n".join(
                sentence.text
                for sentence in collection.sentences
                if not skip_empty_sentences or sentence.keyphrases or sentence.relations
            ),
            encoding="utf8",
        )

    @classmethod
    def _dump_ann(
        cls, collection: Collection, ann_path: Path, skip_empty_sentences=True
    ):
        collection.fix_ids()

        aid = 0
        rid = 0
        shift = 0
        with ann_path.open("w", encoding="utf8") as ann_file:
            for sentence in collection.sentences:
                if (
                    skip_empty_sentences
                    and not sentence.keyphrases
                    and not sentence.relations
                ):
                    continue

                for keyphrase in sentence.keyphrases:
                    ann_file.write(keyphrase.as_ann(shift))

                    for attribute in keyphrase.attributes:
                        ann_file.write(attribute.as_ann(aid))
                        aid += 1

                for relation in sentence.relations:
                    ann_file.write(relation.as_ann(rid))
                    if relation.label != "same-as":
                        rid += 1

                shift += len(sentence) + 1


class DisjointSet:
    def __init__(self, *items):
        self.nodes = {x: DisjointNode(x) for x in items}

    def merge(self, items):
        items = (self.nodes[x] for x in items)
        try:
            head, *others = items
            for other in others:
                head.merge(other)
        except ValueError:
            pass

    @property
    def representatives(self):
        return {n.representative for n in self.nodes.values()}

    @property
    def groups(self):
        return [
            [n for n in self.nodes.values() if n.representative == r]
            for r in self.representatives
        ]

    def __len__(self):
        return len(self.representatives)

    def __getitem__(self, item):
        return self.nodes[item]

    def __call__(self, item1, item2):
        return self[item1].representative == self[item2].representative

    def __str__(self):
        return str(self.groups)

    def __repr__(self):
        return str(self)


class DisjointNode:
    def __init__(self, value):
        self.value = value
        self.parent = self

    @property
    def representative(self):
        if self.parent != self:
            self.parent = self.parent.representative
        return self.parent

    def merge(self, other):
        other.representative.parent = self.representative

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self)



try:
    import streamlit as st
except ImportError:
    print(
        "(!) The code inside `autogoal.contrib.streamlit` requires `streamlit>=0.55`."
    )
    print("(!) Fix it by running `pip install autogoal[streamlit]`.")
    raise


from autogoal.search import Logger


class StreamlitLogger(Logger):
    def __init__(self):
        self.evaluations = 0
        self.current = 0
        self.status = st.info("Waiting for evaluation start.")
        self.progress = st.progress(0)
        self.error_log = st.empty()
        self.best_fn = 0
        self.chart = st.line_chart([dict(current=0.0, best=0.0)])
        self.current_pipeline = st.code("")
        self.best_pipeline = None

    def begin(self, evaluations, pop_size):
        self.status.info(f"Starting evaluation for {evaluations} iterations.")
        self.progress.progress(0)
        self.evaluations = evaluations

    def update_best(self, new_best, new_fn, previous_best, previous_fn):
        self.best_fn = new_fn
        self.best_pipeline = repr(new_best)

    def sample_solution(self, solution):
        self.current += 1
        self.status.info(
            f"""
            [Best={self.best_fn:0.3}] 🕐 Iteration {self.current}/{self.evaluations}.
            """
        )
        self.progress.progress(self.current / self.evaluations)
        self.current_pipeline.code(repr(solution))

    def eval_solution(self, solution, fitness):
        self.chart.add_rows([dict(current=fitness, best=self.best_fn)])

    def end(self, best, best_fn):
        self.status.success(
            f"""
            **Evaluation completed:** 👍 Best solution={best_fn:0.3}
            """
        )
        self.progress.progress(1.0)
        self.current_pipeline.code(self.best_pipeline)



# coding: utf-8

import os
import numpy as np
from scipy import sparse as sp
from autogoal.datasets import datapath, download

def load():
    """
    Loads train and valid datasets from [Gisette uci dataset](https://archive.ics.uci.edu/ml/datasets/Gisette).

    ##### Examples

    ```python
    >>> X_train, y_train, X_valid, y_valid = load()
    >>> X_train.shape, X_valid.shape
    ((6000, 5000), (1000, 5000))
    >>> len(y_train), len(y_valid)
    (6000, 1000)

    ```
    """

    try:
        download("gisette")
    except:
        print("Error loading data. This may be caused due to bad connection. Please delete badly downloaded data and retry")
        raise

    path = str(datapath(os.path.dirname(os.path.abspath(__file__)))) + "/data/gisette"

    train_data = open(os.path.join(path, "gisette_train.data"), "r")
    train_labels = open(os.path.join(path, "gisette_train.labels"), "r")
    valid_data = open(os.path.join(path, "gisette_valid.data"), "r")
    valid_labels = open(os.path.join(path, "gisette_valid.labels"), "r")

    Xtrain = sp.lil_matrix((6000, 5000))
    ytrain = []
    Xvalid = sp.lil_matrix((1000, 5000))
    yvalid = []

    for i, line in enumerate(train_data):
        for j, value in enumerate(line.split()):
            value = int(value)
            if value > 0:
                Xtrain[i,j] = value

    for i, line in enumerate(valid_data):
        for j, value in enumerate(line.split()):
            value = int(value)
            if value > 0:
                Xvalid[i,j] = value

    for line in train_labels:
        ytrain.append(int(line) > 0)

    for line in valid_labels:
        yvalid.append(int(line) > 0)

    return Xtrain.tocsr(), np.asarray(ytrain), Xvalid.tocsr(), np.asarray(yvalid)



from autogoal.contrib.keras._generated import *
from autogoal.grammar import GraphGrammar, Path, Block, CfgInitializer, Epsilon


class Module:
    def make_top_level(self, top_level: list):
        pass

    def add_productions(self, grammar: GraphGrammar):
        raise NotImplementedError()


class Modules:
    class Preprocessing:
        class Recurrent(Module):
            def make_top_level(self, top_level):
                if "PreprocessingModule" not in top_level:
                    top_level.append("PreprocessingModule")
            
            def add_productions(self, grammar: GraphGrammar):
                grammar.add("PreprocessingModule", "RecurrentModule")
                
                grammar.add("RecurrentModule", Path("RecurrentCell", "RecurrentModule"))
                grammar.add("RecurrentModule", "RecurrentCell")

                grammar.add("RecurrentCell", Seq2SeqLSTM)
                grammar.add("RecurrentCell", Seq2SeqBiLSTM)

        class Conv2D(Module): 
            def make_top_level(self, top_level):
                if "PreprocessingModule" not in top_level:
                    top_level.append("PreprocessingModule")

            def add_productions(self, grammar: GraphGrammar):
                grammar.add("PreprocessingModule", Path("Conv2DModule", Flatten))
                
                grammar.add("Conv2DModule", Path("Conv2DBlock", "Conv2DModule"))
                grammar.add("Conv2DModule", "Conv2DBlock")

                grammar.add("Conv2DBlock", Path("Conv2DCells", MaxPooling2D))
                grammar.add("Conv2DBlock", Path("Conv2DCells", MaxPooling2D, Dropout))

                grammar.add("Conv2DCells", Path("Conv2DCell", "Conv2DCells"))
                grammar.add("Conv2DCells", Path("Conv2DCell"))

                grammar.add("Conv2DCell", Path(Conv2D, Activation, BatchNormalization))
                grammar.add("Conv2DCell", Path(Conv2D, Activation))

    class Features:
        class Dense(Module):
            def make_top_level(self, top_level):
                if "FeaturesModule" not in top_level:
                    top_level.append("FeaturesModule")

            def add_productions(self, grammar):
                grammar.add("FeaturesModule", Path("DenseModule", "FeaturesModule"))
                grammar.add("FeaturesModule", Epsilon())

                grammar.add("DenseModule", Block("DenseCell", "DenseModule"))
                grammar.add("DenseModule", Path("DenseCell", "DenseModule"))
                grammar.add("DenseModule", Epsilon())
                grammar.add("DenseCell", Path(Dense, Activation, Dropout))
                grammar.add("DenseCell", Path(Dense, Activation))


def generate_grammar(*modules):
    top_level = []

    for mod in modules:
        mod.make_top_level(top_level)

    grammar = GraphGrammar(start=Path(*top_level), initializer=CfgInitializer())

    for mod in modules:
        mod.add_productions(grammar)

    return grammar


def build_grammar(
    preprocessing=False,
    preprocessing_recurrent=True,
    preprocessing_conv=False,
    reduction=False,
    reduction_recurrent=True,
    reduction_conv=False,
    features=False,
    features_time_distributed=False,
):
    modules = []

    if preprocessing:
        modules.append("PreprocessingModule")

    if reduction:
        modules.append("ReductionModule")

    if features:
        modules.append("FeaturesModule")

    if features_time_distributed:
        modules.append("FeaturesTimeDistributedModule")

    if preprocessing_recurrent and preprocessing_conv:
        raise ValueError("Cannot combine recurrent with convolutional preprocessing.")

    if reduction_recurrent and reduction_conv:
        raise ValueError(
            "Cannot combine recurrent with convolutional reduction modules."
        )

    if (reduction or features) and features_time_distributed:
        raise ValueError("Cannot combine time-distributed modules with flat modules.")

    if not modules:
        raise ValueError("At least one module must be activated.")

    grammar = GraphGrammar(start=Path(*modules), initializer=CfgInitializer())

    if preprocessing_recurrent:
        grammar.add("PreprocessingModule", "PreprocessingModuleR")
        grammar.add("PreprocessingModuleR", Path("Recurrent", "PreprocessingModuleR"))
        grammar.add("PreprocessingModuleR", Epsilon())
        grammar.add("Recurrent", Seq2SeqLSTM)
        grammar.add("Recurrent", Seq2SeqBiLSTM)

    if preprocessing_conv:
        grammar.add(
            "PreprocessingModule", Path("Convolutional", "PreprocessingModuleC")
        )
        grammar.add(
            "PreprocessingModuleC", Path("Convolutional", "PreprocessingModuleC")
        )
        grammar.add("PreprocessingModuleC", Epsilon())
        grammar.add("Convolutional", Path(Conv2D, MaxPooling2D))
        grammar.add("Convolutional", Path(Conv2D, MaxPooling2D, Dropout))

    if reduction_recurrent:
        grammar.add("ReductionModule", Seq2VecLSTM)
        grammar.add("ReductionModule", Seq2VecBiLSTM)

    if reduction_conv:
        grammar.add("ReductionModule", Flatten)

    if features:
        grammar.add("FeaturesModule", Path("DenseLayer", "FeaturesModule"))
        grammar.add("FeaturesModule", Epsilon())
        grammar.add("DenseLayer", Block(Dense, "DenseLayer"))
        grammar.add("DenseLayer", Path(Dense, "DenseLayer"))
        grammar.add("DenseLayer", Path(Dense, Dropout, "DenseLayer"))
        grammar.add("DenseLayer", Epsilon())

    if features_time_distributed:
        grammar.add(
            "FeaturesTimeDistributedModule",
            Path("TDLayer", "FeaturesTimeDistributedModule"),
        )
        grammar.add("FeaturesTimeDistributedModule", Epsilon())
        grammar.add("TDLayer", Block(TimeDistributed, "TDLayer"))
        grammar.add("TDLayer", Path(TimeDistributed, "TDLayer"))
        grammar.add("TDLayer", Epsilon())

    return grammar



from ._base import SearchAlgorithm, ProgressLogger, ConsoleLogger, Logger, MemoryLogger
from ._random import RandomSearch
from ._pge import ModelSampler, PESearch
from ._learning import SurrogateSearch



"""
This module contains wrappers for several estimators and transformers
from `scikit-learn`.

!!! warning
    Importing this module requires `sklearn` with a version equal or greater
    than `0.22`. You can either install it manually or with `pip install autogoal[sklearn]`.

Most of the classes and functions inside this module deal with the automatic
generation of wrappers and thus are considered private API.

The main public functionality exposed by this module is the function
[find_classes](/api/autogoal.contrib.sklearn/#find_classes), which allows to
enumerate the wrappers implemented in this module applying some filters.

!!! note
    You can manually import any wrapper class directly from `autogoal.contrib.sklearn._generated`
    buy beware that namespace changes wildly from version to version and classes in it
    might disappear or change their signature anytime.
"""

try:
    import sklearn

    major, minor, *rest = sklearn.__version__.split(".")
    assert int(major) == 0 and int(minor) >= 22
except:
    print("(!) Code in `autogoal.contrib.sklearn` requires `sklearn=^0.22`.")
    print("(!) You can install it with `pip install autogoal[sklearn]`.")
    raise ImportError()


from ._generated import *
from ._manual import *
